/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"encoding/base64"
	"errors"
	"fmt"
	"io"
	"os"
	"path"
	"path/filepath"
	"strings"
	"syscall"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rphttp"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/ladapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/signals"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"github.com/creack/goselect"
	"github.com/gofrs/uuid"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"golang.org/x/term"
)

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initLocalCmd(rootCmd *cobra.Command) {

	// localCmd represents the local command
	var localCmd = &cobra.Command{
		Use:   localCommand,
		Short: localCmdShortDesc,
	}

	// localExecCmd represents the "exec" command for local
	var localExecCmd = &cobra.Command{
		Use:          localExecCmdName,
		Short:        localExecCmdShortUsage,
		Long:         localExecCmdLongUsage,
		SilenceUsage: true,
		PreRunE:      initEventsRPLAD,
		RunE:         execCmdLine,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE", // No need for backend connection for this command
			"__lad_connect__": "TRUE"},
	}

	// localPkgCmd represents the "local package" command
	var localPkgCmd = &cobra.Command{
		Use:     localPkgCmdName,
		Aliases: []string{localPkgCmdAlias},
		Short:   localPkgCmdShortUsage,
	}

	// localPkgInitCmd represents the "local package init" command
	var localPkgInitCmd = &cobra.Command{
		Use:               localPkgInitCmdName,
		Short:             localPkgInitCmdShortUsage,
		Long:              localPkgInitCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              packageInit,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgRunCmd represents the "local package build" command
	var localPkgRunCmd = &cobra.Command{
		Use:               localPkgBuildCmdName,
		Aliases:           []string{localPkgRunCmdName},
		Short:             localPkgRunCmdShortUsage,
		Long:              localPkgRunCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              packageRun,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgCleanCmd represents the "local package clean" command
	var localPkgCleanCmd = &cobra.Command{
		Use:               localPkgCleanCmdName,
		Short:             localPkgCleanCmdShortUsage,
		Long:              localPkgCleanCmdLongUsage,
		SilenceUsage:      true,
		RunE:              packageClean,
		ValidArgsFunction: getListAppsFromBldDirAsTab,
	}

	// connectionsCmd represents the "local connections" command
	var connectionsCmd = &cobra.Command{
		Use:   connectionsCommand,
		Short: localConnectionsCmdShortDesc,
	}

	// connectionsListCmd represents the "local connections list" command
	var connectionsListCmd = &cobra.Command{
		Use:         connectionsListCmd,
		Short:       fmt.Sprintf(connectionsListCmdShortDesc, "local-builders"),
		Long:        fmt.Sprintf(connectionsListCmdLongDesc, "local-builders"),
		RunE:        listConnections,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// getCmd represents the get command
	var connectionsGetCmd = &cobra.Command{
		Use:               connectionsGetCmd,
		Short:             connectionsGetCmdShortUsage,
		RunE:              getConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// connectionsCurrentCmd represents the current command
	var connectionsCurrentCmd = &cobra.Command{
		Use:         connectionsCurrentCmd,
		Short:       connectionsCurrentCmdShortUsage,
		RunE:        currentConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// addCmd represents the get command
	var connectionsAddCmd = &cobra.Command{
		Use:         connectionsAddCmd,
		Short:       connectionsAddShortDesc,
		Long:        fmt.Sprintf(connectionsAddLongDesc, "local-builders"),
		RunE:        addConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// updateCmd represents the update command
	var connectionsUpdateCmd = &cobra.Command{
		Use:               connectionsUpdateCmd,
		Short:             connectionsUpdateShortDesc,
		RunE:              updateConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// deleteCmd represents the delete command
	var connectionsDeleteCmd = &cobra.Command{
		Use:               connectionsDeleteCmd,
		Short:             fmt.Sprintf(connectionsDeleteCmdShortUsage, "local-builder"),
		RunE:              deleteConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// useCmd represents the use command
	var connectionsUseCmd = &cobra.Command{
		Use:               connectionsUseCmd,
		Short:             fmt.Sprintf(connectionsUseCmdShortDesc, "local-builder"),
		RunE:              useConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	localPkgCmd.PersistentFlags().StringP(localPkgCmdSourceFlag, localPkgCmdSourceFlagAlias, "./", localPkgCmdSourceFlagUsage)
	localPkgCmd.PersistentFlags().StringP(localPkgCmdPackageFlag, localPkgCmdPackageFlagAlias, "./../gitpkgs/", localPkgCmdPackageFlagUsage)
	// ----------------------------------------------------------

	// Bind the cobra flags to viper
	viper.BindPFlag(localPkgCmdSourceFlag, localPkgCmd.PersistentFlags().Lookup(localPkgCmdSourceFlag))   // Corresponding environment var: RP_SOURCESDIR
	viper.BindPFlag(localPkgCmdPackageFlag, localPkgCmd.PersistentFlags().Lookup(localPkgCmdPackageFlag)) // Corresponding environment var: RP_PKGSDIR
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Exec
	localExecCmd.Flags().StringP(localExecCmdWorkingDirFlag, localExecCmdWorkingDirFlagAlias, "", localExecCmdWorkingDirFlagUsage)
	// Package init
	localPkgInitCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	localPkgInitCmd.Flags().String(localPkgInitSpecPathFlag, "", localPkgInitSpecPathFlagUsage)
	localPkgInitCmd.Flags().String(localPkgInitSpecLocalFlag, "", localPkgInitSpecLocalFlagUsage)
	localPkgInitCmd.Flags().String(localPkgInitSrcURLFlag, "", localPkgInitSrcURLFlagUsage)
	localPkgInitCmd.Flags().String(localPkgInitSrcRevFlag, "", localPkgInitSrcRevFlagUsage)
	localPkgInitCmd.Flags().String(localPkgInitPkgURLFlag, "", localPkgInitPkgURLFlagUsage)
	localPkgInitCmd.Flags().Bool(localPkgInitOfflineFlag, false, localPkgInitOfflineFlagUsage)
	// Package run
	localPkgRunCmd.Flags().String(localPkgRunProjectFlag, "", localPkgRunProjectFlagUsage)
	localPkgRunCmd.Flags().Bool(localPkgRunCmdRegenArchFlag, false, localPkgRunCmdRegenArchFlagUsage)
	localPkgRunCmd.Flags().String(localPkgRunTargetFlag, "aarch64", localPkgRunTargetFlagUsage)
	localPkgRunCmd.Flags().BoolP(localPkgRunCmdNoDelFlag, localPkgRunCmdNoDelFlagAlias, false, localPkgRunCmdNoDelFlagUsage)
	localPkgRunCmd.Flags().Bool(localPkgRunCmdNoCleanAfterFlag, false, localPkgRunCmdNoCleanAfterFlagUsage)
	localPkgRunCmd.Flags().String(localPkgRunCmdInstallFlag, "", localPkgRunCmdInstallFlagUsage)
	localPkgRunCmd.Flags().Bool(localPkgRunCmdShellFlag, false, localPkgRunCmdShellFlagUsage)
	localPkgRunCmd.Flags().String(localPkgRunForceReleaseFlag, "", localPkgRunForceReleaseFlagUsage)
	localPkgRunCmd.Flags().StringP(localPkgRunCmdDistroFlag, localPkgRunCmdDistroFlagAlias, "", localPkgRunCmdDistroFlagUsage)
	// Package clean
	localPkgCleanCmd.Flags().BoolP(localPkgCleanInitFlag, localPkgCleanInitFlagAlias, false, localPkgCleanInitFlagUsage)
	localPkgCleanCmd.Flags().String(localPkgCleanProjectFlag, "", localPkgCleanProjectFlagUsage)
	localPkgCleanCmd.Flags().BoolP(localPkgCleanForceFlag, localPkgCleanForceFlagAlias, false, localPkgCleanForceFlagUsage)
	// Connections list
	connectionsListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections get
	connectionsGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections current
	connectionsCurrentCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections add
	connectionsAddCmd.Flags().StringP(connectionsAddIPFlag, connectionsAddIPFlagAlias, "", connectionsAddIPFlagUsage)
	// Connections update
	connectionsUpdateCmd.Flags().StringP(connectionsUpdateIPFlag, connectionsUpdateIPFlagAlias, flagUpdateKeepCurrentValue, connectionsUpdateIPFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required

	// Add some flags as hidden
	localPkgRunCmd.Flags().MarkHidden(localPkgRunForceReleaseFlag)

	// Specify dynamic flag completion
	localPkgInitCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgRunCmd.RegisterFlagCompletionFunc(localPkgRunProjectFlag, getListPrjsAsTab)
	localPkgRunCmd.RegisterFlagCompletionFunc(localPkgRunTargetFlag, getListArchAsTab)
	localPkgCleanCmd.RegisterFlagCompletionFunc(localPkgCleanProjectFlag, getListPrjsFromBldDirAsTab)

	// Add the command "local" as a subcommand of rp-cli
	rootCmd.AddCommand(localCmd)
	// Add the command "exec" as a subcommand of localCmd
	localCmd.AddCommand(localExecCmd)
	// Add the command "package" as a subcommand of localCmd
	localCmd.AddCommand(localPkgCmd)
	// Add the command "connections" as a subcommand of localCmd
	localCmd.AddCommand(connectionsCmd)
	// Add the command "init" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgInitCmd)
	// Add the command "run" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgRunCmd)
	// Add the command "clean" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgCleanCmd)
	// Add the command "list" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsListCmd)
	// Add the command "get" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsGetCmd)
	// Add the command "current" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsCurrentCmd)
	// Add the command "add" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsAddCmd)
	// Add the command "update" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsUpdateCmd)
	// Add the command "delete" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsDeleteCmd)
	// Add the command "use" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsUseCmd)
}

// initEventsRPLAD inits everything that is needed to communicate
// with the Redpesk LAD service through the websocket
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventsRPLAD(cmd *cobra.Command, args []string) error {

	// Subscribe to all the events in relation with local-builder
	err := localBConnection.RPEventRegister(localStdOutEvt)
	if err != nil {
		return err
	}
	err = localBConnection.RPEventRegister(localStdErrEvt)
	if err != nil {
		return err
	}
	err = localBConnection.RPEventRegister(localProcEndEvt)
	if err != nil {
		return err
	}

	// Save the call-backs needed when some events are received
	// from the local builder
	err = localBConnection.WSocket.OnEvent(localStdOutEvt, CLIReceiveLocalBEvent)
	if err != nil {
		return err
	}
	err = localBConnection.WSocket.OnEvent(localStdErrEvt, CLIReceiveLocalBEvent)
	if err != nil {
		return err
	}
	err = localBConnection.WSocket.OnEvent(localProcEndEvt, CLIReceiveLocalBEvent)
	if err != nil {
		return err
	}

	return nil

}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListAppsFromBldDirAsTab gets all the applications slugs
// corresponding to the directories name used to build locally
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListAppsFromBldDirAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}

	// Try to get the packages directory from Viper
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// Now construct the glob expression to find the right directories
	tmpPkgsDir = path.Join(tmpPkgsDir, "*", "*")

	// Now find every paths matching the glob expression
	pkgDirectories, err := filepath.Glob(tmpPkgsDir)
	if err != nil {
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// For every path found, get only the last part of the path (i.e application slug)
	for _, tmpPath := range pkgDirectories {
		isDir, err := utils.IsDirectory(tmpPath)
		if err != nil {
			return appsList, cobra.ShellCompDirectiveNoFileComp
		}

		if isDir {
			appsList = append(appsList, path.Base(tmpPath))
		}
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getListPrjsFromBldDirAsTab gets all the applications slugs
// corresponding to the directories name used to build locally
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListPrjsFromBldDirAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	prjsList := []string{}

	// Try to get the packages directory from Viper
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return prjsList, cobra.ShellCompDirectiveNoFileComp
	}

	// Now construct the glob expression to find the right directories
	tmpPkgsDir = path.Join(tmpPkgsDir, "*")

	// Now find every paths matching the glob expression
	pkgDirectories, err := filepath.Glob(tmpPkgsDir)
	if err != nil {
		return prjsList, cobra.ShellCompDirectiveNoFileComp
	}

	// For every path found, get only the last part of the path (i.e application slug)
	for _, tmpPath := range pkgDirectories {
		isDir, err := utils.IsDirectory(tmpPath)
		if err != nil {
			return prjsList, cobra.ShellCompDirectiveNoFileComp
		}

		baseName := path.Base(tmpPath)

		if isDir && baseName != localPrjNotSpecified {
			prjsList = append(prjsList, baseName)
		}
	}

	return prjsList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// execCmdLine runs the command line inside the local-builder
// configured on the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func execCmdLine(cmd *cobra.Command, args []string) error {

	// Get the working directory to use (if set by user)
	workingDir, err := cmd.Flags().GetString(localExecCmdWorkingDirFlag)
	if err != nil {
		logFlagError(localExecCmdWorkingDirFlag)
		return err
	}

	// If the user did not set any working directory, check if we are not inside
	// a gitsources/ or gitpkgs/ directory or sub-directory
	if workingDir == "" {
		// Get the "pkgsdir" value in config file
		pkgsDir := ""
		pkgsDir, err = manuallyGetPkgDirFlag()
		if err != nil {
			return err
		}
		if pkgsDir != "" {
			pkgsDir, err = utils.SanitizePath(pkgsDir)
			if err != nil {
				return err
			}
		}

		// Get the "sourcesdir" value in config file
		srcDir := ""
		srcDir, err = manuallyGetSrcDirFlag()
		if err != nil {
			return err
		}
		if srcDir != "" {
			srcDir, err = utils.SanitizePath(srcDir)
			if err != nil {
				return err
			}
		}

		// Get the current working directory
		currentDir, err := os.Getwd()
		if err != nil {
			return err
		}

		// Check if the user is in gitpkgs/ dir
		isInPkgsDir := strings.HasPrefix(currentDir, pkgsDir)

		// Check if the user is in gitsources/ dir
		isInSrcsDir := strings.HasPrefix(currentDir, srcDir)

		if isInPkgsDir && (pkgsDir != "") {
			// Replace the right directory in the workingDir to pass
			workingDir = strings.ReplaceAll(currentDir, pkgsDir, localBuilderGitPkgsDir)
		} else if isInSrcsDir && (srcDir != "") {
			// Replace the right directory in the workingDir to pass
			workingDir = strings.ReplaceAll(currentDir, srcDir, localBuilderGitSourcesDir)
		}
	}

	// Get the command line passed by the user
	cmdLine, err := getCmdLine(args)
	if err != nil {
		return err
	}

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Request local-builder to run the command
	// Request body
	execRequest := ladapi.RPLADExecCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: workingDir,
		},
		CmdLine: cmdLine,
	}
	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.ExecPost(&execRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)

	return err
}

// packageInit initializes everything that is needed to package
// an application inside the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageInit(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine0)

	// Before anything check that 'git' and 'git-lfs' are installed in the system
	_, err := utils.RunCmdInDirectory("/tmp/", "git", "--version")
	if err != nil {
		return errors.New(gitPkgNotInstalled)
	}
	_, err = utils.RunCmdInDirectory("/tmp/", "git", "lfs", "--version")
	if err != nil {
		return errors.New(gitLFSPkgNotInstalled)
	}

	// Get the first main infos we need
	fullConfiguration, appStructure, err := prePackagingRoutine(cmd, args, true)
	if err != nil {
		return err
	}

	// Display something like "Retrieving packaging configurations" here
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine5, appStructure.Slug)

	// From everything we know about the app, retrieve the structure
	// containing all the data needed to set-up the local builder
	// (spec file link, GIT URL, revisions, etc.)
	err = getUserBuildSetUp(cmd, &fullConfiguration)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something like "Set-upping the local builder" here
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine6)

	// Set-up the local builder
	err = localBuilderSetUp(&appStructure, &fullConfiguration)
	if err != nil {
		return err
	}

	return nil
}

// packageRun run the packaging of the application passed as argument
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageRun(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localPkgRunCmdAnsLine0)

	// Get the first main infos we need
	fullConfiguration, appStructure, err := prePackagingRoutine(cmd, args, false)
	if err != nil {
		return err
	}

	// Display something like "Retrieving packaging configurations" here
	RPPrinter.CustomPrintF(localPkgRunCmdAnsLine3, appStructure.Slug)

	// Retrieve the target tag value from the given project
	targetTag, err := getAppTargetTag(cmd, &appStructure, &fullConfiguration)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Get the architecture wanted by the user in order to run "localbuild" or "localcrossbuild"
	targetWanted, err := cmd.Flags().GetString(localPkgRunTargetFlag)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Download or retrieve the right configuration file (according online/offline status)
	buildConfigFile, err := getAppBuildConfigFile(cmd, appStructure.Parent, &fullConfiguration, targetWanted)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Get the regen archive flag to see if we need to recreate the tarball
	regenTar, err := cmd.Flags().GetBool(localPkgRunCmdRegenArchFlag)
	if err != nil {
		return err
	}
	if regenTar && (fullConfiguration.PackageURL == "") {
		RPPrinter.CustomPrintLn(localPkgRunCmdAnsLine1)
		err = regenTarball(&fullConfiguration, appStructure.PackageName)
		if err != nil {
			return err
		}
	} else if regenTar && (fullConfiguration.PackageURL != "") {
		RPPrinter.CustomColorPrintLn(printer.ColorYellow, localPkgRunCmdAnsLine8)
		RPPrinter.CustomPrintLn()
	}

	// Run the appropriate command in the local builder
	RPPrinter.CustomPrintLn(localPkgRunCmdAnsLine2)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgRunCmdAnsPkg1)
	RPPrinter.ForceFlush()
	var errToReturn error
	var isTruePackaging bool
	if targetWanted == "aarch64" {
		isTruePackaging, errToReturn = runIotpkgLocalCrossbuild(cmd, &fullConfiguration, targetTag, targetWanted, buildConfigFile)
	} else if targetWanted == "x86_64" {
		isTruePackaging, errToReturn = runIotpkgLocalbuild(cmd, &fullConfiguration, targetTag, buildConfigFile)
	} else {
		return errors.New(targetNotSupportedErr)
	}
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgRunCmdAnsPkg2)

	// If it is a true packaging command (no shell or things like that),
	// tell where the packages are (on the host)
	if isTruePackaging {
		// Find the specfile path
		specFilePath := ""
		if !fullConfiguration.IsSpecExt && !fullConfiguration.IsSpecLocal && fullConfiguration.SpecFilePath != "" {
			specFilePath = path.Join(fullConfiguration.SourcesDir, fullConfiguration.SpecFilePath)
		} else {
			specFilesTab, _ := utils.FindFilesWithExtension(fullConfiguration.PackagesDir, ".spec")
			if len(specFilesTab) != 1 {
				RPPrinter.CustomColorPrintF(printer.ColorYellow, localPkgRunCmdAnsLine9, len(specFilesTab), fullConfiguration.PackagesDir)
				RPPrinter.CustomPrintLn()
			} else {
				specFilePath = specFilesTab[0]
			}
		}

		if specFilePath != "" {
			// Get the spec file version
			specVersion, err := utils.GetSpecFileVersion(specFilePath)
			if err != nil {
				return err
			}

			// Get the spec file release
			specRelease, err := utils.GetSpecFileRelease(specFilePath)
			if err != nil {
				return err
			}

			// Construct the directory containing the results
			resultsPath := fmt.Sprintf("%s/results_%s/%s/%s.rp*/", fullConfiguration.PackagesDir, appStructure.Slug, specVersion, specRelease)

			// Display the info to the user
			RPPrinter.CustomColorPrintLn(printer.ColorCyan, localPkgRunCmdAnsLine4, resultsPath)
			RPPrinter.CustomPrintLn()
		}

		// Congrats the user if the packaging went well!
		if errToReturn == nil {
			RPPrinter.CustomColorPrintLn(printer.ColorGreen, localPkgRunCmdAnsLine5)
			RPPrinter.CustomPrintLn()
			// Suggest a command to create the application on Redpesk
			// if it does not already exists
			if appStructure.ID == uuid.Nil {
				// Create the command line
				cmdLine := createAddAppCmd(&fullConfiguration, &appStructure)
				// Display info to the user
				RPPrinter.CustomPrintLn(localPkgRunCmdAnsLine7)
				RPPrinter.CustomPrintLn()
				RPPrinter.CustomPrintLn("> " + cmdLine)
				RPPrinter.CustomPrintLn()
			}
		} else {
			RPPrinter.CustomColorPrintLn(printer.ColorRed, localPkgRunCmdAnsLine6)
		}
	}

	return errToReturn
}

// packageClean cleans the repositories used to build locally an
// application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageClean(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localPkgCleanCmdAnsLine0)
	RPPrinter.CustomPrintLn()

	// Initialize the configuration structure that we may use
	fullConfiguration := RPLocalBuildConfig{}

	// Before anything check that 'git' is installed in the system
	_, err := utils.RunCmdInDirectory("/tmp/", "git", "--version")
	if err != nil {
		return errors.New(gitPkgNotInstalled)
	}

	// Get the init flag
	initFlag, err := cmd.Flags().GetBool(localPkgCleanInitFlag)
	if err != nil {
		logFlagError(localPkgCleanInitFlag)
		return err
	}

	// Get the force flag
	forceFlag, err := cmd.Flags().GetBool(localPkgCleanForceFlag)
	if err != nil {
		logFlagError(localPkgCleanForceFlag)
		return err
	}

	// Try to find the paths of the directories to delete
	srcDir, pkgsDir, err := findPathsToDelete(cmd, args)
	if err != nil {
		return err
	}

	if initFlag {
		// Retrieve the configuration file from the already existing pkgs directory
		configFilePath := path.Join(pkgsDir, localCfgSubDirectory, "local_build_config.json")
		err = utils.ReadJsonFile(configFilePath, &fullConfiguration)
		if err != nil {
			return err
		}
	}

	if !forceFlag && srcDir != "" {
		// Check if there is modification in source directory
		gitStatusOutput, err := utils.RunCmdInDirectory(srcDir, "git", "status")
		if err != nil {
			Log.Warnf(noGitSourceFoundWarn, gitStatusOutput)
		} else if !strings.Contains(gitStatusOutput, "nothing to commit, working tree clean") {
			// Ask the user if he really wants to delete the repository
			answerLow := rpYesOrNoQuestion(localPkgCleanCmdAnsLine1)
			if answerLow == "n" {
				// The user does not want to delete the directory, stop here
				return nil
			}
		}
	}

	// Delete the mock if requested
	deleteMock := false
	if cmd.Flags().Lookup(localV2CleanNoMockFlag) != nil {
		tmpNoDelMock, err := cmd.Flags().GetBool(localV2CleanNoMockFlag)
		if err != nil {
			return err
		}
		deleteMock = !tmpNoDelMock
	}
	if deleteMock {
		iotpkgFullCfgDir := path.Join(pkgsDir, localIotpkgSubDirectory)

		// Check if the .iotpkg/ directory exists
		_, err := os.Stat(iotpkgFullCfgDir)
		if !os.IsNotExist(err) {
			// Get the name of the mock(s) to delete (go look into .iotpkg/ directory inside the gitpkgs/ directory)
			dirsAndFilesMap, err := utils.FindAllFilesAndDirsRecursively(iotpkgFullCfgDir)
			if err != nil {
				return err
			}

			// For each mock name found, delete the mock and the corresponding ".cfg" file
			for _, filepath := range dirsAndFilesMap["files"] {
				tmpFilename := path.Base(filepath)

				// Find the real mock name within the .cfg file
				mockName, err := getMockNameFromCfgFile(filepath)
				if err != nil {
					Log.Errorf("error while getting the mock name for '%s': %+v", tmpFilename, err)
				}

				// Check that we got a name
				if mockName == "" {
					Log.Warnf("No mock found for %s", tmpFilename)
					continue
				}

				// User communication
				RPPrinter.CustomPrintF("Deleting mock '%s'...\n", mockName)

				// Ask redpesk LAD to delete the mock
				err = handleIotpkgCleanMock(mockName)
				if err != nil {
					Log.Errorf("error while cleaning the mock named '%s': %+v", mockName, err)
					// Does not stop on error at this level
				}

				// Now, delete the corresponding .cfg file
				tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
				if err != nil {
					return err
				}
				err = os.Remove(path.Join(tmpPkgsDir, tmpFilename))
				if err != nil {
					return err
				}
			}
		}
	}

	// Delete the directories
	err = os.RemoveAll(srcDir)
	if err != nil {
		return err
	}
	err = os.RemoveAll(pkgsDir)
	if err != nil {
		return err
	}

	// If the parent directory is empty, delete it as well!
	parentSrcDir := path.Dir(srcDir)
	err = utils.RemoveDirIfEmpty(parentSrcDir)
	if err != nil {
		return err
	}
	parentPkgDir := path.Dir(pkgsDir)
	err = utils.RemoveDirIfEmpty(parentPkgDir)
	if err != nil {
		return err
	}

	if initFlag {
		// Run the local package init command using the fields retrieved in the configuration file

		// Get the root command
		rpCliRootCmd := cmd.Root()

		// Find the local package init command from that
		cmdLine := []string{"local", "package", "init", path.Base(srcDir)}
		localPkgInitCmd, inArgs, err := rpCliRootCmd.Find(cmdLine)
		if err != nil {
			return err
		}

		projectSlug := path.Base(path.Dir(srcDir))

		// Set the different flags needed
		localPkgInitCmd.Flags().Set(localPkgCmdPackageFlag, fullConfiguration.PackagesDir)
		localPkgInitCmd.Flags().Set(localPkgCmdSourceFlag, fullConfiguration.SourcesDir)

		if fullConfiguration.OfflineMode {
			localPkgInitCmd.Flags().Set(localPkgInitOfflineFlag, "true")
		}

		if fullConfiguration.PackageURL != "" {
			localPkgInitCmd.Flags().Set(localPkgInitPkgURLFlag, fullConfiguration.PackageURL)
		}

		if projectSlug != localPrjNotSpecified {
			localPkgInitCmd.Flags().Set(localPkgInitProjectFlag, projectSlug)
		}

		if fullConfiguration.SourcesRev != "" {
			localPkgInitCmd.Flags().Set(localPkgInitSrcRevFlag, fullConfiguration.SourcesRev)
		}

		if fullConfiguration.SourcesURL != "" {
			localPkgInitCmd.Flags().Set(localPkgInitSrcURLFlag, fullConfiguration.SourcesURL)
		}

		if fullConfiguration.IsSpecLocal && fullConfiguration.SpecFilePath != "" {
			localPkgInitCmd.Flags().Set(localPkgInitSpecLocalFlag, fullConfiguration.SpecFilePath)
		} else if fullConfiguration.SpecFilePath != "" {
			localPkgInitCmd.Flags().Set(localPkgInitSpecPathFlag, fullConfiguration.SpecFilePath)
		}

		// As log info, print the command line we are about to play
		// Log.Info(initAfterCleanInfoMsg, strings.Join(cmdLine, " "))

		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()

		// Now call the command
		err = localPkgInitCmd.PreRunE(localPkgInitCmd, inArgs)
		if err != nil {
			return err
		}

		err = localPkgInitCmd.RunE(localPkgInitCmd, inArgs)
		if err != nil {
			return err
		}
	}

	RPPrinter.CustomColorPrintLn(printer.ColorGreen, localPkgCleanCmdAnsLine2)

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getCmdLine get the command line furnished by the user. If no
// command is passed or if it is do not respect the format, it
// returns an error
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: string, error -- the command line and an error if something went wrong
func getCmdLine(args []string) (string, error) {
	// Initialize the result
	resultCmd := ""

	// Check that the command is at least 1 args long ('--' is not counted)
	if len(args) < 1 {
		return resultCmd, errors.New(cmdToShortError)
	}

	return strings.Join(args, " "), nil
}

// prePackagingRoutine is a common "routine" between `package init` and
// `package build`. It allows to retrieve all the information that are common
// to `run` and `init` that have been set by the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param isInit: Tell if the function is called at init or at run time
// :type isInit: bool
// :returns: RPLocalBuildConfig, RPApplicationFull, error -- build config, the local app structure
// and an error if something went wrong
func prePackagingRoutine(cmd *cobra.Command, args []string, isInit bool) (RPLocalBuildConfig, rpapiv1.RPApplicationFull, error) {

	// Initialize the main structures we need to return
	fullConfiguration := RPLocalBuildConfig{}

	// Display something like "Checking application details" here
	RPPrinter.CustomPrint(localPkgInitCmdAnsLine1)

	// Get the name of the application that will be packaged
	// and if the application exists
	appStructure, err := getLocalApp(cmd, args)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return fullConfiguration, appStructure, err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something like "Checking packaging directories" here
	RPPrinter.CustomPrint(localPkgInitCmdAnsLine2)

	// Get and check the 'gitsources/' and 'gitpkgs/' directory
	dirSuffix := ""
	if !isInit {
		dirSuffix = path.Join(appStructure.Parent.ProjectSlug, appStructure.Slug)
	}
	tmpSrcsDir, tmpPkgsDir, err := getAndChecksBuildDirs(dirSuffix)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return fullConfiguration, appStructure, err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	if !isInit {

		// We are not at initialization, get all the info from the config file
		configFilePath := path.Join(tmpPkgsDir, localCfgSubDirectory, "local_build_config.json")
		err = utils.ReadJsonFile(configFilePath, &fullConfiguration)
		if err != nil {
			return fullConfiguration, appStructure, err
		}

	} else {
		// We are at initialization
		// Set what we can do from the app
		fullConfiguration.SourcesRev = appStructure.Metadata.Data.Revision
		fullConfiguration.SourcesURL = appStructure.Metadata.Data.SourcesURI
		fullConfiguration.SpecFilePath = appStructure.Metadata.Data.SpecfileLocation
	}

	// Save the packages and sources directories
	fullConfiguration.SourcesDir = tmpSrcsDir
	fullConfiguration.PackagesDir = tmpPkgsDir

	// Display the path to both directories here
	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine3, fullConfiguration.SourcesDir)
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine4, fullConfiguration.PackagesDir)
	RPPrinter.CustomPrintLn()

	return fullConfiguration, appStructure, nil
}

// getLocalApp parses the user command line and retrieve the
// application UUID and slug. It also permits to check that the corresponding
// project is created and if the application exists in Redpesk
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: rpapiv1.RPApplicationFull, error -- the application (empty struct if not existsing) and an error if something went wrong
func getLocalApp(cmd *cobra.Command, args []string) (rpapiv1.RPApplicationFull, error) {
	// Set the default values
	appExists := true

	app, err := genericGetAppFromArgs(cmd, args, true)
	if err != nil {
		// Only keep working with error, if the error is about the application not existing
		if !strings.HasPrefix(err.Error(), "no application has been found with") {
			return rpapiv1.RPApplicationFull{}, err
		} else {
			appExists = false
		}
	}

	// If the app exists, get the full structure
	if !appExists {
		// Since the application does not exists, get what has been passed by the user
		tmpSlug := ""
		tmpSlug, err = getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "application"), fmt.Sprintf(objectIDRequiredMsg, "application"))
		if err != nil {
			return rpapiv1.RPApplicationFull{}, err
		}

		// Sanitize what the user gave us
		app.Slug, err = utils.SanitizeAppName(tmpSlug)
		if err != nil {
			return rpapiv1.RPApplicationFull{}, err
		}

		// Set-up the same name for the package name
		app.PackageName = app.Slug

		// Save the project slug
		appPrjName, err := cmd.Flags().GetString(localPkgRunProjectFlag)
		if err != nil {
			return rpapiv1.RPApplicationFull{}, err
		}
		app.Parent.ProjectSlug = appPrjName

		// Set a default project name if empty
		if app.Parent.ProjectSlug == "" {
			app.Parent.ProjectSlug = localPrjNotSpecified
		}

	}

	return app, nil
}

// getCustomBuildsDirs is the function that looks for the different builds directories inside
// viper environment but also in the connection parameters if the user set-up a custom path.
// The priority is given to the custom value
//
// :param flagName: Flag to look for
// :type flagName: string
// :returns: string, error -- path to the directory and an error if something went wrong
func getCustomBuildsDirs(flagName string) (string, error) {
	// First thing to do, check if the flag is defined in the connection file for the current server alias
	// Get the alias of the remote connection
	serverAlias := viper.GetString(serverAliasFlag)

	// Get the connection parameters
	connecParams, err := getConnectionParams(serverAlias, accessTokenFile)
	if err != nil {
		return "", err
	}

	// Look for the right field in the connection
	strToConsider := ""
	if flagName == localPkgCmdSourceFlag {
		strToConsider = connecParams.SourcesDir
	} else if flagName == localPkgCmdPackageFlag {
		strToConsider = connecParams.PkgsDir
	}

	if strToConsider != "" {
		return strToConsider, nil
	}

	// If we reach here, let's return the default value
	return viper.GetString(flagName), nil
}

// getAndChecksBuildDirs get the gitsources/ and gitpkgs/
// directories on the host sides. It checks them as well
//
// :param dirSuffix: Additional dir to add at the end of directory to check (if not empty)
// :type dirSuffix: string
// :returns: string, string, error -- sources dir, pkgs dir and an error if something went wrong
func getAndChecksBuildDirs(dirSuffix string) (string, string, error) {
	// Try to get the sources directory from everywhere we can
	tmpSrcDir, err := getCustomBuildsDirs(localPkgCmdSourceFlag)
	if err != nil {
		return "", "", err
	}

	// Try to get the packages directory from Viper
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return "", "", err
	}

	// Add suffix if set in parameter
	checkSourceDir := true
	if dirSuffix != "" {
		tmpSrcDir = path.Join(tmpSrcDir, dirSuffix)
		tmpPkgsDir = path.Join(tmpPkgsDir, dirSuffix)
		// If a suffix has been set, we are at running time!
		// No need to check for source directory (it is possible that it is not existing)
		checkSourceDir = false
	}

	// SOURCES DIRECTORY --------------
	var SrcDir string

	// Sanitize path (~ and ./, ../, etc.)
	SrcDir, err = utils.SanitizePath(tmpSrcDir)
	if err != nil {
		return "", "", err
	}

	// Check that the resulting directory exists
	if _, err := os.Stat(SrcDir); checkSourceDir && os.IsNotExist(err) {
		return "", "", err
	}
	// --------------------------------

	// PACKAGES DIRECTORY -------------
	var PkgsDir string

	// Sanitize path (~ and ./, ../, etc.)
	PkgsDir, err = utils.SanitizePath(tmpPkgsDir)
	if err != nil {
		return "", "", err
	}

	// Check that the resulting directory exists
	if _, err := os.Stat(PkgsDir); os.IsNotExist(err) {
		return "", "", err
	}
	// --------------------------------

	return SrcDir, PkgsDir, nil
}

// findPathsToDelete tries to look into the local packaging directories in order to find
// one that match the application slug and optionally the project slug given in arguments
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: string, string, error -- sources dir, pkgs dir and an error if something went wrong
func findPathsToDelete(cmd *cobra.Command, args []string) (string, string, error) {
	// Try to get the sources directory from everywhere we can
	tmpSrcDir, err := getCustomBuildsDirs(localPkgCmdSourceFlag)
	if err != nil {
		return "", "", err
	}

	// Try to get the packages directory from Viper
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return "", "", err
	}

	// From command line get the application details
	appStructure, err := getLocalAppV2(cmd, args)
	if err != nil {
		return "", "", err
	}

	// Now construct the paths to the right directories
	srcDirectory := path.Join(tmpSrcDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)
	pkgDirectory := path.Join(tmpPkgsDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Check that the source directory exists
	_, err = os.Stat(srcDirectory)
	if os.IsNotExist(err) {
		Log.Warnf(localCleanNoSourcesDirWarn, appStructure.Slug)
		srcDirectory = ""
	}

	// Check that the gitpkg directory exists
	_, err = os.Stat(pkgDirectory)
	if os.IsNotExist(err) {
		return "", "", fmt.Errorf(packageCleanDirNotFound, pkgDirectory)
	}

	return srcDirectory, pkgDirectory, nil
}

// getUserBuildSetUp retrieves the value given by the user.
// If the application already exists and the user gives additional
// values, the user values are used.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param app: Structure containing all needed data for application
// :type app: *RPLocalApplication
// :param configToFill: Configuration to fill with the right values
// :type configToFill: *RPLocalBuildConfig
// :returns: error -- an error if something went wrong
func getUserBuildSetUp(cmd *cobra.Command, configToFill *RPLocalBuildConfig) error {

	// Retrieve what the user has furnished

	// Get the URL of the sources, if the user specified it
	appSrcURL, err := cmd.Flags().GetString(localPkgInitSrcURLFlag)
	if err != nil {
		return err
	}
	if appSrcURL != "" {
		configToFill.SourcesURL = appSrcURL
	}

	// Get the URL of the package that could have been passed
	appPkgURL, err := cmd.Flags().GetString(localPkgInitPkgURLFlag)
	if err != nil {
		return err
	}
	if appPkgURL != "" {
		configToFill.PackageURL = appPkgURL
	}

	// Get the revision of the source, if the user specified it
	appSrcRev, err := cmd.Flags().GetString(localPkgInitSrcRevFlag)
	if err != nil {
		return err
	}
	if appSrcRev != "" {
		configToFill.SourcesRev = appSrcRev
	}

	// Get the path to the spec file, if the user specified it
	appSpecPath, err := cmd.Flags().GetString(localPkgInitSpecPathFlag)
	if err != nil {
		return err
	}
	if appSpecPath != "" {
		configToFill.SpecFilePath = appSpecPath
	}

	// Get the path to the local spec file, if the user specified it
	appSpecLocal, err := cmd.Flags().GetString(localPkgInitSpecLocalFlag)
	if err != nil {
		return err
	}
	if appSpecLocal != "" {
		// Sanitize the path to the spec file before saving it
		configToFill.SpecFilePath, err = utils.SanitizePath(appSpecLocal)
		if err != nil {
			return err
		}
		configToFill.IsSpecLocal = true
	}

	// Get if we are in offline mode or not (if this flag exists)
	if cmd.Flags().Lookup(localPkgInitOfflineFlag) != nil {
		modeOffline, err := cmd.Flags().GetBool(localPkgInitOfflineFlag)
		if err != nil {
			return err
		}
		configToFill.OfflineMode = modeOffline
	}

	// Check that the configurations are coherent

	// Check that we have at least an URL for the sources or a package
	if (configToFill.SourcesURL == "") && (configToFill.PackageURL == "") {
		return errors.New(noSourcesURLError)
	}
	// Check that we have only one URL (sources or package)
	if (configToFill.SourcesURL != "") && (configToFill.PackageURL != "") {
		return errors.New(duplicateSrcURL)
	}
	// If empty, put the revision to "master"
	if configToFill.SourcesRev == "" {
		configToFill.SourcesRev = "master"
	}
	if configToFill.IsSpecLocal {
		// Check that the specfile exists locally
		if _, err := os.Stat(configToFill.SpecFilePath); os.IsNotExist(err) {
			return errors.New(noLocalSpecFileError)
		}
	}

	return nil
}

// localBuilderSetUp fills the RPLocalBuildConfig structure with the values left
// to set-up fully the local-builder to build. It manages both cases where the
// application already exists in the Redpesk factory, or not.
//
// :param app: Structure containing all needed data for application
// :type app: *rpapiv1.RPApplicationFull
// :param config: Configuration to set for the local builder
// :type config: *RPLocalBuildConfig
// :returns: error -- an error if something went wrong
func localBuilderSetUp(app *rpapiv1.RPApplicationFull, localConfig *RPLocalBuildConfig) error {
	// Create the path to the directories that are interesting us ------------------
	// Project directories
	ourSourcesPrjDir := path.Join(localConfig.SourcesDir, app.Parent.ProjectSlug)
	ourPackagesPrjDir := path.Join(localConfig.PackagesDir, app.Parent.ProjectSlug)
	// App directories
	ourSourcesDir := path.Join(ourSourcesPrjDir, app.Slug)
	ourPackagesDir := path.Join(ourPackagesPrjDir, app.Slug)
	// rp-cli configuration directories
	ourCfgPkgDir := path.Join(ourPackagesDir, localCfgSubDirectory)
	// -----------------------------------------------------------------------------

	// Create the project level directories if there are not existing
	err := utils.CreateDirIfNotExists(ourSourcesPrjDir, false)
	if err != nil {
		return err
	}
	err = utils.CreateDirIfNotExists(ourPackagesPrjDir, false)
	if err != nil {
		return err
	}

	// Return an error if a similar project already exists in gitsources/ or gitpkgs/
	if _, err := os.Stat(ourSourcesDir); !os.IsNotExist(err) {
		// The sources directory already exists, request to delete it
		return fmt.Errorf(dirExistsError, ourSourcesDir)
	}
	if _, err := os.Stat(ourPackagesDir); !os.IsNotExist(err) {
		// The sources directory already exists, request to delete it
		return fmt.Errorf(dirExistsError, ourPackagesDir)
	}

	// Check if the spec file is located in another GIT repo
	localConfig.IsSpecExt = false
	if !localConfig.IsSpecLocal &&
		(strings.HasPrefix(localConfig.SpecFilePath, "http:") || strings.HasPrefix(localConfig.SpecFilePath, "https:")) {
		localConfig.IsSpecExt = true
	}

	// Now that everything is checked, set-up the local builder

	// Save the repo to clone and where to clone it!
	var repoToClone string
	var dirForGit string
	if localConfig.PackageURL != "" {
		repoToClone = localConfig.PackageURL
		dirForGit = ourPackagesPrjDir
	} else {
		repoToClone = localConfig.SourcesURL
		dirForGit = ourSourcesPrjDir
	}

	// Clone the sources inside gitsources/
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine7)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgInitCmdAnsClone1)
	RPPrinter.ForceFlush()
	gitCloneOutput, err := utils.RunCmdInDirectory(dirForGit, "git", "clone", "--branch", localConfig.SourcesRev, "--progress", repoToClone, app.Slug)
	RPPrinter.CustomPrint(gitCloneOutput)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgInitCmdAnsClone2)
	RPPrinter.ForceFlush()
	if err != nil {
		return err
	}

	// If the sources has been cloned, create the package directory
	// In the other case the git-package has been cloned directly!

	if localConfig.PackageURL == "" {
		// Create the packages directory
		// Do a `git init` to make`iotpkg localbuild` happy
		// FIXME: it would be cool to have iotpkg to not needs that!
		// err = os.MkdirAll(ourPackagesDir, 0755)
		// if err != nil {
		// 	return err
		// }

		_, err = utils.RunCmdInDirectory(ourPackagesPrjDir, "git", "init", app.Slug)
		if err != nil {
			return err
		}
		_, err = utils.RunCmdInDirectory(ourPackagesDir, "git", "remote", "add", "origin", localConfig.SourcesURL)
		if err != nil {
			return err
		}
	}

	// Create the ".rp-cli" configuration sub-directory if needed
	err = utils.CreateDirIfNotExists(ourCfgPkgDir, false)
	if err != nil {
		return err
	}

	// Retrieve the build configuration file if we are in "Offline mode"
	if localConfig.OfflineMode {
		// Retrieve all the build configuration files
		RPPrinter.CustomPrint(localPkgInitCmdAnsLine13)
		err = createAllBuildConfigFiles(app.Parent, ourCfgPkgDir)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

		// Retrieve all the released tags that can possibly exists
		RPPrinter.CustomPrint(localPkgInitCmdAnsLine14)
		allReleasedTags, err := getAllTargetTagsFromPrj(app.Parent)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		localConfig.ReleasedTags = allReleasedTags
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}

	// Write a file in the package directory to save all the configuration
	configFilePath := path.Join(ourCfgPkgDir, "local_build_config.json")
	err = utils.WriteJsonFile(configFilePath, localConfig)
	if err != nil {
		return err
	}

	if localConfig.PackageURL == "" {

		// Working out the spec file
		RPPrinter.CustomPrint(localPkgInitCmdAnsLine8)
		specOnHost, err := manageSpecFileSetUp(localConfig, app, ourSourcesDir, ourPackagesDir)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

		// Create a tarball from the sources
		RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine11)
		RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgInitCmdAnsArchive1)
		RPPrinter.ForceFlush()
		err = createSrcTarball(localConfig, ourPackagesDir, specOnHost)
		if err != nil {
			return err
		}
		RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgInitCmdAnsArchive2)

		// If the spec file was local or external tell the user to directly modify it in gitpkgs/<app>/
		if localConfig.IsSpecLocal || localConfig.IsSpecExt || (localConfig.SpecFilePath == "") {
			specFileName := filepath.Base(specOnHost)
			RPPrinter.CustomColorPrintF(printer.ColorCyan, localPkgInitCmdAnsLine9, specFileName, ourPackagesDir)
			RPPrinter.CustomColorPrintF(printer.ColorCyan, localPkgInitCmdAnsLine10, path.Join(ourPackagesDir, specFileName))
			RPPrinter.CustomPrintLn()
		}
	} else {
		// Find the spec file cloned from the "git-package"
		specFiles, err := utils.FindFilesWithExtension(ourPackagesDir, ".spec")
		if err != nil {
			return err
		}

		if len(specFiles) == 1 {
			RPPrinter.CustomColorPrintF(printer.ColorCyan, localPkgInitCmdAnsLine10, specFiles[0])
		} else {
			RPPrinter.CustomColorPrintF(printer.ColorCyan, localPkgInitCmdAnsLine12, specFiles[0])
		}

		RPPrinter.CustomPrintLn()
	}

	return nil
}

// manageSpecFileSetUp allows to set-up the dpec file in the different use cases:
// 1/ The spec file is local (furnished by the user)
// 2/ The spec file is in the sources repository
// 3/ The spec file is in another repo
//
// :param config: Configuration to set for the local builder
// :type config: *RPLocalBuildConfig
// :param app: Application parameters needed to set-up the environment
// :type app: *rpapiv1.RPApplicationFull
// :param appSrcDir: Directory containing the source files of the app
// :type appSrcDir: string
// :param appPkgsDir: Directory containing the items needed to package the app
// :type appPkgsDir: string
// :returns: string, error -- the path to the spec file on the host and an error if something went wrong
func manageSpecFileSetUp(config *RPLocalBuildConfig, app *rpapiv1.RPApplicationFull, appSrcDir, appPkgsDir string) (string, error) {
	var err error

	specFilePathHost := ""
	appExists := (app.ID != uuid.Nil)

	// Check that we have a spec file path if the application does not exists
	if (config.SpecFilePath == "") && !appExists {
		return specFilePathHost, errors.New(noSpecPathError)
	}

	// In any case we can get the spec file name
	specFileName := ""
	if config.SpecFilePath != "" {
		specFileName = filepath.Base(config.SpecFilePath)
	} else {
		specFileName = app.PackageName + ".spec"
	}

	// Switch between the different scenarios for the spec file
	switch {
	case config.IsSpecLocal:

		specFilePathHost = path.Join(appPkgsDir, specFileName)

		// If the spec file is local, directly copy it in the package directory
		err = utils.CopyFile(config.SpecFilePath, specFilePathHost)

	case config.IsSpecExt:
		// If the spec file is in another GIT repo, download the file in gitpkgs/<app>/

		specFilePathHost = path.Join(appPkgsDir, specFileName)

		err = rphttp.HTTPDownloadFile(specFilePathHost, config.SpecFilePath)
		// TODO: another management smarter than that (i.e. cloning the GIT repo containing the spec)?

	case (config.SpecFilePath != ""):
		// If the spec file is inside the sources, create a symlink in gitpkgs/<app>/

		specFilePathHost = path.Join(appSrcDir, config.SpecFilePath)

		// Construct the symbolic link path in local builder
		symlink := path.Join(localBuilderGitPkgsDir, app.Parent.ProjectSlug, path.Base(appPkgsDir), specFileName)
		target := path.Join(localBuilderGitSourcesDir, app.Parent.ProjectSlug, path.Base(appPkgsDir), config.SpecFilePath)

		// Add debug logs
		Log.Debugf(symlinkReqToLad, target, symlink)

		// Request redpesk-lad to create the symbolic link
		err = localBAPI.CreateSymlink(symlink, target)

	case (config.SpecFilePath == "") && appExists:
		// If the spec file has been directly uploaded in Redpesk factory, download it from it

		specFilePathHost = path.Join(appPkgsDir, specFileName)

		// Download the file from the Redpesk factory
		rpFile := rpapiv1.RPFile_Deprecated{}
		err = bckAPI.ApplicationGetFile(app.ID.String(), specFileName, &rpFile)
		if err != nil {
			break
		}

		// Write the file to gitpkgs/<app>/
		err = utils.WriteToFile(specFilePathHost, rpFile.Content)

	default:
		return specFilePathHost, errors.New(specScenarioNotManagedErr)
	}

	return specFilePathHost, err
}

// createSrcTarball allows to request the local-builder to create the
// source archive in the right directory.
//
// :param config: Full configuration of the local builder
// :type config: *RPLocalBuildConfig
// :param appPkgsDir: Directory containing the items needed to package the app
// :type appPkgsDir: string
// :param specFilePath: Path to the spec file on the host (needed to get its version)
// :type specFilePath: string
// :returns: error -- an error if something went wrong
func createSrcTarball(config *RPLocalBuildConfig, appPkgsDir, specFilePath string) error {
	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Get the name of the directory for our application
	appDirName := path.Base(appPkgsDir)
	// Get the name of the parent directory
	prjDirName := path.Base(path.Dir(appPkgsDir))

	// Create the path of the directories within the local builder
	localBuildSrcDir := path.Join(localBuilderGitSourcesDir, prjDirName, appDirName)
	localBuildPkgDir := path.Join(localBuilderGitPkgsDir, prjDirName, appDirName)

	// Get the current version inside the spec file and force the version
	// to this one in `iotpkg archive`
	specVersion, err := utils.GetSpecFileVersion(specFilePath)
	if err != nil {
		return err
	}

	// Request iotpkg to be verbose
	iotpkgFlags := map[string]string{
		"-v": "",
	}

	// Set the right flags for the archive subcommand
	archiveFlags := map[string]string{
		"--path-source": localBuildSrcDir,
		"--commit":      config.SourcesRev,
		"--version":     specVersion,
		"--no-specfile": "",
	}

	// Declare the body of the request to send to redpesk-lad server
	archiveRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBuildPkgDir,
		},
		IotpkgFlags: iotpkgFlags,
		SubcmdFlags: archiveFlags,
	}

	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgArchivePost(&archiveRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)

	return err
}

// regenTarball allows to regen the archive of a sources directory already
// present in local builder. This function can be use by `rp-cli package build`
//
// :param config: Full configuration of the local builder
// :type config: *RPLocalBuildConfig
// :param pkgName: Name of the source package
// :type pkgName: string
// :returns: error -- an error if something went wrong
func regenTarball(config *RPLocalBuildConfig, pkgName string) error {

	// Create the spec file path from the data we have
	specFile := path.Join(config.PackagesDir, pkgName+".spec")

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgInitCmdAnsArchive1)
	RPPrinter.ForceFlush()

	// Call the right function to regenerate the archive
	err := createSrcTarball(config, config.PackagesDir, specFile)
	if err != nil {
		return err
	}

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localPkgInitCmdAnsArchive2)

	return nil
}

// runIotpkgLocalbuild run the `iotpkg localbuild` inside the
// local builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param config: Full configuration of the local builder
// :type config: *RPLocalBuildConfig
// :param release: Release to use for the packaging
// :type release: string
// :param buildConfigFile: Configuration file name used to set-up the mock
// :type buildConfigFile: string
// :returns: bool, error -- true if it is a real package build and an error if something went wrong
func runIotpkgLocalbuild(cmd *cobra.Command, config *RPLocalBuildConfig, release, buildConfigFile string) (bool, error) {

	// Initialize the boolean result
	isTruePackaging := false

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return isTruePackaging, err
	}

	// Get the name of the directory for our application
	appDirName := path.Base(config.PackagesDir)
	// Get the name of the parent directory
	prjDirName := path.Base(path.Dir(config.PackagesDir))

	// Create the path of the package directory within the local builder
	localBuildPkgDir := path.Join(localBuilderGitPkgsDir, prjDirName, appDirName)

	// Construct the full path to the build configuration file
	fullBuildConfigFile := path.Join(localBuildPkgDir, localCfgSubDirectory, buildConfigFile)

	// Retrieve the option that can be passed to iotpkg in localbuilder
	localbuildFlags, err := setIotpkgLocalbuildFlags(cmd)
	if err != nil {
		return isTruePackaging, err
	}

	// Add the "mock-config" flag
	localbuildFlags["--mock-config"] = fullBuildConfigFile

	// Create the map containing the flags for iotpkg
	iotpkgFlags := map[string]string{
		"-v":        "",
		"--release": release,
	}

	// Check if it is a full package command or some other debug commands
	// like "--shell" or "--install"
	_, shellEnabled := iotpkgFlags["--shell"]
	_, installFlagEnabled := iotpkgFlags["--install"]
	if !shellEnabled && !installFlagEnabled {
		isTruePackaging = true
	}

	// Declare the body of the request to send to redpesk-lad server
	localbuildRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBuildPkgDir,
		},
		IotpkgFlags: iotpkgFlags,
		SubcmdFlags: localbuildFlags,
	}

	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgLocalbuildPost(&localbuildRequest, &procTicket)
	if err != nil {
		return isTruePackaging, err
	}

	err = listenStdoutErrAndIn(procTicket)

	return isTruePackaging, err
}

// runIotpkgLocalCrossbuild run the `iotpkg localcrossbuild` inside the
// local builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param config: Full configuration of the local builder
// :type config: *RPLocalBuildConfig
// :param release: Release to use for the packaging
// :type release: string
// :param arch: Target architecture for the packaging
// :type arch: string
// :param buildConfigFile: Configuration file name used to set-up the mock
// :type buildConfigFile: string
// :returns: bool, error -- true if it is a real package build and an error if something went wrong
func runIotpkgLocalCrossbuild(cmd *cobra.Command, config *RPLocalBuildConfig, release, arch, buildConfigFile string) (bool, error) {

	// Initialize the boolean result
	isTruePackaging := false

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return isTruePackaging, err
	}

	// Get the name of the directory for our application
	appDirName := path.Base(config.PackagesDir)
	// Get the name of the parent directory
	prjDirName := path.Base(path.Dir(config.PackagesDir))

	// Create the path of the package directory within the local builder
	localBuildPkgDir := path.Join(localBuilderGitPkgsDir, prjDirName, appDirName)

	// Construct the full path to the build configuration file
	fullBuildConfigFile := path.Join(localBuildPkgDir, localCfgSubDirectory, buildConfigFile)

	// Retrieve the option that can be passed to iotpkg in localbuilder
	localbuildCrossFlags, err := setIotpkgLocalbuildFlags(cmd)
	if err != nil {
		return isTruePackaging, err
	}

	// Add the arch flag
	localbuildCrossFlags["--arch"] = arch

	// Add the "mock-config" flag
	localbuildCrossFlags["--mock-config"] = fullBuildConfigFile

	// Create the map containing the flags for iotpkg
	iotpkgFlags := map[string]string{
		"-v":        "",
		"--release": release,
	}

	// Check if it is a full package command or some other debug commands
	// like "--shell" or "--install"
	_, shellEnabled := iotpkgFlags["--shell"]
	_, installFlagEnabled := iotpkgFlags["--install"]
	if !shellEnabled && !installFlagEnabled {
		isTruePackaging = true
	}

	// Declare the body of the request to send to redpesk-lad server
	localbuildRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBuildPkgDir,
		},
		IotpkgFlags: iotpkgFlags,
		SubcmdFlags: localbuildCrossFlags,
	}

	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgLocalcrossbuildPost(&localbuildRequest, &procTicket)
	if err != nil {
		return isTruePackaging, err
	}

	err = listenStdoutErrAndIn(procTicket)

	return isTruePackaging, err
}

// setIotpkgLocalbuildFlags creates the map of string containing
// the flags necessary for iotpkg localbuild or localcrossbuild
// from the flags that the user gave us.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: map[string]string, error -- the map and an error if something went wrong
func setIotpkgLocalbuildFlags(cmd *cobra.Command) (map[string]string, error) {

	// Initialization
	result := make(map[string]string)

	// Get the no-delete flag
	noDelete, err := cmd.Flags().GetBool(localPkgRunCmdNoDelFlag)
	if err != nil {
		return result, err
	}

	// Get the no-clean-after flag
	noCleanAfter, err := cmd.Flags().GetBool(localPkgRunCmdNoCleanAfterFlag)
	if err != nil {
		return result, err
	}

	// Get the shell flag
	shellFlag, err := cmd.Flags().GetBool(localPkgRunCmdShellFlag)
	if err != nil {
		return result, err
	}

	// Get the install flag
	installFlag, err := cmd.Flags().GetString(localPkgRunCmdInstallFlag)
	if err != nil {
		return result, err
	}

	// Set-up the map from what we have
	if noDelete {
		result["-N"] = ""
	}

	if noCleanAfter {
		result["--no-cleanup-after"] = ""
	}

	if shellFlag {
		result["--shell"] = ""
	}

	if installFlag != "" {
		result["--install"] = installFlag
	}

	return result, nil
}

// listenStdoutErrAndIn allows to listen simultaneously stdin from the
// user and stdout/err from the remote running process.
// Basically, this function is really handly to call afetr sending a
// command to run on the remote server. This command could be interactive
// or not.
//
// :param procTicket: The ticket received from the server and identifying the
// remote process
// :type procTicket: ladapi.RPLADRunnerTicket
// :returns: error -- an error if something went wrong, nil in the other case
func listenStdoutErrAndIn(procTicket ladapi.RPLADRunnerTicket) error {

	// Setup terminal (raw mode to handle escape and control keys)
	if (!term.IsTerminal(int(os.Stdin.Fd())) || !term.IsTerminal(int(os.Stdout.Fd()))) &&
		(runMode != "testing") {
		return errors.New(stdShouldBeTerminal)
	}
	oldState, err := term.MakeRaw(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}
	defer term.Restore(int(os.Stdin.Fd()), oldState)

	// Start the process listening on stdin
	exitChannel := liveListenStdin(procTicket.ID, oldState)

	// Handle signals
	err = signals.OnSignals(func(sig os.Signal) {
		Log.Debugf("Send signal %v", sig)

		if signals.IsWinResizeSignal(sig) {
			// Request Redpesk LAD to resize the terminal size
			terminalResizeWin(procTicket.ID)

		} else if signals.IsInterruptSignal(sig) {
			stdinEvt := RPLocalBStdinEvent{
				ID:    procTicket.ID,
				Chars: "\x04\n",
			}
			localBConnection.WSocket.EmitEvent(localStdInEvt, &stdinEvt)

		} else {
			// Send the signal to the server (and the process)
			terminalSendSignal(procTicket.ID, sig)
		}
	})
	if err != nil {
		return err
	}

	// Block until we receive all the websocket events (until localProcEndEvt)
	procRunning := true
	var errToReturn error
	for procRunning {
		select {
		// We received an event from the remote server
		case localEvent := <-cmdEventLocalBChannel:
			// Check that the received event is for our process
			if localEvent.ID == procTicket.ID {
				switch localEvent.Type {
				case localStdOutEvt:
					// Print to the os.Stdout if the line is not empty
					if len(localEvent.Line) > 0 {
						os.Stdout.Write([]byte(localEvent.Line))
					}
				case localStdErrEvt:
					// Print to the os.Stderr if the line is not empty
					if len(localEvent.Line) > 0 {
						os.Stderr.Write([]byte(localEvent.Line))
					}
				case localProcEndEvt:
					// This is the end of the terminal
					errorMsg := fmt.Sprintf(endOfProc, localEvent.ID, localEvent.Status)
					Log.Debugln(errorMsg)

					// End this blocking code
					procRunning = false

					// Return an error if the terminal end badly
					if localEvent.Status != 0 {
						errToReturn = errors.New(errorMsg)
					} else {
						errToReturn = nil
					}
				}
			}

		// End of terminal (high chance for error, see liveListenStdin())
		case res := <-exitChannel:
			// End this blocking code
			procRunning = false
			errToReturn = nil

			if res.Error != nil {
				Log.Errorf(execErrorMsg, res.Error)
				errToReturn = res.Error
			}

		// Disconnection from the websocket, since everything go through the
		// socket, stop the command
		case <-localBConnection.WSocket.DisconnectionChan:
			// End this blocking code
			procRunning = false
			Log.Error(webSocketDisconnectionErr)
			errToReturn = errors.New(webSocketDisconnectionErr)
		}
	}

	return errToReturn
}

// liveListenStdin is a function to call as a coroutine in order to
// listen on the stdin of the terminal. It is used to send stdin event
// to the local-builder server.
//
// :param id: ID of the process to send event to
// :type id: string
// :param oldState: Previous state of the terminal
// :type oldState: *term.State
// :returns: chan RPLocalExitResult -- the channel to use to stop the process
func liveListenStdin(id string, oldState *term.State) chan RPLocalExitResult {
	exitChannel := make(chan RPLocalExitResult)

	go func(id string, oldState *term.State) {
		// Declaration of exposeFd type
		type exposeFd interface {
			Fd() uintptr
		}

		buff := make([]byte, 128)
		rdfs := &goselect.FDSet{}
		reader := io.ReadCloser(os.Stdin)
		defer reader.Close()

		for {
			rdfs.Zero()
			rdfs.Set(reader.(exposeFd).Fd())
			err := goselect.Select(int(os.Stdin.Fd())+1, rdfs, nil, nil, 50*time.Millisecond)
			if err != nil {
				if err == syscall.EINTR {
					continue
				} else {
					term.Restore(int(os.Stdin.Fd()), oldState)
					exitChannel <- RPLocalExitResult{err, 3}
					return
				}
			}
			if rdfs.IsSet(reader.(exposeFd).Fd()) {
				size, err := reader.Read(buff)

				if err != nil {

					Log.Debugf("Read error %v; err %v", size, err)
					if err == io.EOF {

						// CTRL-D exited scanner, so send it explicitly
						stdinEvt := RPLocalBStdinEvent{
							ID:    id,
							Chars: "\x04\n",
						}
						err := localBConnection.WSocket.EmitEvent(localStdInEvt, &stdinEvt)

						// If a problem occurs, restore the old terminal
						if err != nil {
							term.Restore(int(os.Stdin.Fd()), oldState)
							exitChannel <- RPLocalExitResult{err, 4}
							return
						}

						time.Sleep(time.Millisecond * 100)
						continue

					} else {
						term.Restore(int(os.Stdin.Fd()), oldState)
						exitChannel <- RPLocalExitResult{err, 5}
						return
					}
				}

				if size <= 0 {
					continue
				}

				// Nominal way: send the input to the server
				data := buff[:size]
				stdinEvt := RPLocalBStdinEvent{
					ID:    id,
					Chars: string(data),
				}
				err = localBConnection.WSocket.EmitEvent(localStdInEvt, &stdinEvt)

				// If a problem occurs, restore the old terminal
				if err != nil {
					term.Restore(int(os.Stdin.Fd()), oldState)
					exitChannel <- RPLocalExitResult{err, 6}
					return
				}
			}
		}

	}(id, oldState)

	return exitChannel
}

// runCmdInLocalBuilderWithoutLogs is a function to call locally in this package.
// It allows to run a command inside the local builder without displaying any logs
// like stdout or stderr. It only return how the command ended.
//
// :param commandLine: Command line to play inside the local builder
// :type commandLine: string
// :param workingDir: Directory where the command needs to be run (can be empty)
// :type workingDir: string
// :param timeoutSec: Timeout of the command, in seconds
// :type timeoutSec: int
// :returns: error -- error representing how the command line ended
func runCmdInLocalBuilderWithoutLogs(commandLine, workingDir string, timeoutSec int) error {
	// Create the channel we will use
	timeOutChannel := make(chan bool)

	// Variable containing the loop condition
	procRunning := true

	// Request local-builder to run the command
	// Request body
	execRequest := ladapi.RPLADExecCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: 100,
				WinSizeCol:  100,
			},
			WorkingDir: workingDir,
		},
		CmdLine: commandLine,
	}

	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err := localBAPI.ExecPost(&execRequest, &procTicket)
	if err != nil {
		return err
	}

	// Create the Goroutine waiting until time-out
	go func(timeout int, runningCond *bool) {
		// Count until timeout
		counter := 0
		for (counter < timeoutSec) && *runningCond {
			time.Sleep(1 * time.Second)
			counter++
		}

		// If we reach the time-out send something on the channel
		if *runningCond {
			timeOutChannel <- true
		}
	}(timeoutSec, &procRunning)

	// Wait for the end of the command or for the timeout
	var errToReturn error
	for procRunning {
		select {

		// We received an event from the remote server
		case localEvent := <-cmdEventLocalBChannel:
			// Check that the received event is for our process
			if (localEvent.ID == procTicket.ID) && (localEvent.Type == localProcEndEvt) {
				// This is the end of the terminal
				errorMsg := fmt.Sprintf(endOfProc, localEvent.ID, localEvent.Status)
				Log.Debugln(errorMsg)

				// End this blocking code
				procRunning = false

				// Return an error if the terminal end badly
				if localEvent.Status != 0 {
					errToReturn = errors.New(errorMsg)
				} else {
					errToReturn = nil
				}
			}

		// We reached the timeout
		case <-timeOutChannel:
			// End this blocking code
			procRunning = false
			errorMsg := fmt.Sprintf(endOfProcStrStatus, procTicket.ID, "TIME-OUT")
			errToReturn = errors.New(errorMsg)

		// Disconnection from the websocket, since everything go through the
		// socket, stop the command
		case <-localBConnection.WSocket.DisconnectionChan:
			// End this blocking code
			procRunning = false
			Log.Error(webSocketDisconnectionErr)
			errToReturn = errors.New(webSocketDisconnectionErr)
		}
	}

	return errToReturn
}

// terminalSendSignal allows to send a signal to a process running remotely
// in the container where redpesk LAD is running
//
// :param id: ID of the process to send the signal to
// :type id: string
// :param signal: Signal to send to the process
// :type signal: os.Signal
func terminalSendSignal(id string, signal os.Signal) {
	// Initialize the structure to send to the server
	sigStructure := ladapi.RPLADSignalSend{
		ID:     id,
		Signal: signal.String(),
	}

	// Send the request to the server
	localBAPI.TerminalSendSignalPost(&sigStructure)
}

// terminalResizeWin allows to resize the window of the terminal
// open in the remote server (redpesk LAD)
//
// :param id: ID of the process to resize
// :type id: string
func terminalResizeWin(id string) {
	// Get the current terminal size
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		Log.Errorf(getTermSizeError, err)
	}

	// Initialize the structure to send to the server
	winSizeStructure := ladapi.RPLADWinResize{
		ID: id,
		RPLADWinSize: ladapi.RPLADWinSize{
			WinSizeRows: termHeight,
			WinSizeCol:  termWidth,
		},
	}

	// Send the request to the server
	localBAPI.TerminalSizePut(&winSizeStructure)
}

// getAppTargetTag allows to retrieve the target tag needed to build an
// application from the app it self or from it's corresponding project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param app: Pointer to the application retrieved from what the user gave
// :type app: *rpapiv1.RPApplicationFull
// :param fullConfig: Pointer to the full configuration for the build
// :type fullConfig: *RPLocalBuildConfig
// :returns: string, error -- the release and an error if something went wrong
func getAppTargetTag(cmd *cobra.Command, app *rpapiv1.RPApplicationFull, fullConfig *RPLocalBuildConfig) (string, error) {

	// If the user has forced a release (through the hidden flag), take it and go!
	forcedRelease, err := cmd.Flags().GetString(localPkgRunForceReleaseFlag)
	if err != nil {
		return "", err
	}
	if forcedRelease != "" {
		return forcedRelease, nil
	}

	// Get the distribution flag
	distroWanted, err := cmd.Flags().GetString(localPkgRunCmdDistroFlag)
	if err != nil {
		return "", err
	}

	targetTag := ""
	switch {
	case (fullConfig.OfflineMode && distroWanted == ""):
		// In this case, the user has not specified any distro, and he wants "offline" mode
		// The number of distro inside the saved configuration must be 1!
		if len(fullConfig.ReleasedTags) > 1 {
			errMsg := fmt.Sprintf(needToSpecifyDistroErr, app.Parent.ProjectSlug)
			return "", errors.New(errMsg)
		}

		// Get the only value from the map
		targetTag, err = utils.GetFirstMapValue(fullConfig.ReleasedTags)
		if err != nil {
			return "", err
		}

	case (fullConfig.OfflineMode && distroWanted != ""):
		// In this case, the user has specified a distro, and he wants "offline" mode
		// The distro wanted needs to exist in the saved configuration
		var distroExists bool
		if targetTag, distroExists = fullConfig.ReleasedTags[distroWanted]; !distroExists {
			errMsg := fmt.Sprintf(needToChangeDistroErr, distroWanted, app.Parent.ProjectSlug)
			return "", errors.New(errMsg)
		}

	default:
		// In this case the user does not want an offline mode

		var IDToLookFor string
		if app.ID != uuid.Nil {
			// If the application exists, get the project ID from the struct
			IDToLookFor = app.Parent.ProjectID.String()
		} else {
			// From what the user gave us, get the UUID
			IDToLookFor = app.Parent.ProjectSlug
		}

		// Get the project
		project, err := getProjectFromIdentifier(IDToLookFor, true)
		if err != nil {
			return "", err
		}

		// If no distribution is provided, check that only one distribution exists for this project
		if distroWanted == "" {
			// Get the list of all the "set" distributions
			distroList := []string{}
			for _, tmpDistro := range project.Distributions {
				if tmpDistro.Set {
					distroList = append(distroList, tmpDistro.Slug)
				}
			}

			if len(distroList) > 1 {
				errMsg := fmt.Sprintf(needToSpecifyDistroErr, project.Slug)
				return "", errors.New(errMsg)
			}
			distroWanted = distroList[0]
		}

		// Get the corresponding "target tag"
		targetTag, err = getTargetTagFromPrj(&project, distroWanted)
		if err != nil {
			return "", err
		}
	}

	return targetTag, nil
}

// getAppBuildConfigFile allows to retrieve the build config file needed to package an
// application. This file actually contains the configuration of the mock that will be
// set-up in the local-builder.
//
// This function either choose the right "already downloaded" file if we are working offline
// or download the right one if we are working online.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param appParent: Parent of the application that the user wants to build
// :type appParent: rpapiv1.RPAppParent
// :param fullConfig: Pointer to the full configuration for the build
// :type fullConfig: *RPLocalBuildConfig
// :param arch: Target architecture wanted
// :type arch: string
// :returns: string, error -- the configuration file name and an error if something went wrong
func getAppBuildConfigFile(cmd *cobra.Command, appParent rpapiv1.RPAppParent, fullConfig *RPLocalBuildConfig, arch string) (string, error) {

	// Get the distribution flag
	distroWanted, err := cmd.Flags().GetString(localPkgRunCmdDistroFlag)
	if err != nil {
		return "", err
	}

	// Construct the path to the directory containing the configuration
	ourCfgPkgDir := path.Join(fullConfig.PackagesDir, localCfgSubDirectory)

	// Get the list of possible configuration files in gitpkgs/ directory
	cfgFiles, err := utils.FindFilesWithExtension(ourCfgPkgDir, ".cfg")
	if err != nil {
		return "", err
	}

	configFilePath := ""
	switch {
	case (fullConfig.OfflineMode && distroWanted == ""):
		// In this case, the user has not specified any distro, and he wants "offline" mode
		// The number of available configuration file with the right architecture needs to be 1!

		// Get the number of file corresponding to our arch, and retrieve the match
		regularExpressionWanted := fmt.Sprintf(`_%s.cfg$`, arch)
		nbOfMatch, matchingStr, err := utils.RegexMatchNumberInSlice(regularExpressionWanted, cfgFiles)
		if err != nil {
			return "", err
		}

		if nbOfMatch > 1 {
			errMsg := fmt.Sprintf(needToSpecifyDistroErr, appParent.ProjectSlug)
			return "", errors.New(errMsg)
		} else if nbOfMatch <= 0 {
			errMsg := fmt.Sprintf(NoConfFileFoundErr, mainNotSetMsg, arch)
			return "", errors.New(errMsg)
		}

		// Save the configuration file path
		configFilePath = matchingStr[0]

	case (fullConfig.OfflineMode && distroWanted != ""):
		// In this case, the user has specified a distro, and he wants "offline" mode
		// The configuration file with the right distribution and architecture needs to exists!

		// Get the number of file corresponding to our arch and distro, and retrieve the match
		regularExpressionWanted := fmt.Sprintf(`_%s_%s.cfg$`, distroWanted, arch)
		nbOfMatch, matchingStr, err := utils.RegexMatchNumberInSlice(regularExpressionWanted, cfgFiles)
		if err != nil {
			return "", err
		}

		if nbOfMatch > 1 {
			return "", errors.New(tooManyBuildConfFileFoundErr)
		} else if nbOfMatch <= 0 {
			errMsg := fmt.Sprintf(NoConfFileFoundErr, distroWanted, arch)
			return "", errors.New(errMsg)
		}

		// Save the configuration file path
		configFilePath = matchingStr[0]

	default:
		// In this case the user does not want an offline mode

		// Firstly retrieve the right project UUID
		var prjIDToPass string
		if utils.UUIDEqual(uuid.Nil, appParent.ProjectID) {
			prjIDToPass = appParent.ProjectSlug
		} else {
			prjIDToPass = appParent.ProjectID.String()
		}

		// Retrieve all the arch and distributions from the project
		prjArchDist, prjUUID, err := getDistAndArchFromPrj(prjIDToPass)
		if err != nil {
			return "", err
		}

		// Make a list of all arches and all distros
		allDistros := append(prjArchDist.MandDistros, prjArchDist.OptDistros...)

		// If no distribution is provided, check that only one distribution exists for this project
		if distroWanted == "" {
			if len(allDistros) > 1 {
				errMsg := fmt.Sprintf(needToSpecifyDistroErr, appParent.ProjectSlug)
				return "", errors.New(errMsg)
			}
			distroWanted = allDistros[0]
		}

		// Create the build config file and retrieve its path
		configFilePath, err = createBuildConfigFile(prjUUID, distroWanted, arch, ourCfgPkgDir)
		if err != nil {
			return "", err
		}
	}

	// Now we have the config file path, return only the filename
	return path.Base(configFilePath), nil
}

// createAddAppCmd creates the appropriate rp-cli command
// allowing to add an application on Redpesk after a
// successful local packaging of an application
//
// :param config: Full configuration of the local builder
// :type config: *RPLocalBuildConfig
// :param app: Application structure we want to create
// :type app: *rpapiv1.RPApplicationFull
// :returns: string -- the command line to run
func createAddAppCmd(config *RPLocalBuildConfig, app *rpapiv1.RPApplicationFull) string {

	// Check in which spec file case we are
	specFileTypeValue := ""
	specFileLocationValue := ""

	switch {
	case config.IsSpecExt:
		specFileTypeValue = string(rpapiv1.RPAppSpecTypeExternal)
		specFileLocationValue = config.SpecFilePath
	case config.IsSpecLocal:
		specFileTypeValue = string(rpapiv1.RPAppSpecTypeInternal)
		specFileLocationValue = path.Join(config.PackagesDir, app.PackageName+".spec")
	default:
		specFileTypeValue = string(rpapiv1.RPAppSpecTypeRelative)
		specFileLocationValue = config.SpecFilePath
	}

	// Create the command line
	cmdLine := fmt.Sprintf(localPkgRunAddAppTemplate, app.Slug, app.PackageName, config.SourcesURL, config.SourcesRev,
		app.Parent.ProjectSlug, specFileTypeValue, specFileLocationValue)

	return cmdLine
}

// createAllBuildConfigFiles retrieves the build configuration file from
// the backend and write it in the given target directory
//
// :param appParent: Parent of the application that the user wants to build
// :type appParent: rpapiv1.RPAppParent
// :param targetDir: Directory where the caller wants to create the file
// :type targetDir: string
// :returns: error -- an error if a problem occurred
func createAllBuildConfigFiles(appParent rpapiv1.RPAppParent, targetDir string) error {
	var err error
	var prjIDToPass string

	// If we do not have an UUID for the parent project, we need to retrieve it
	if utils.UUIDEqual(uuid.Nil, appParent.ProjectID) {
		prjIDToPass = appParent.ProjectSlug
	} else {
		prjIDToPass = appParent.ProjectID.String()
	}

	// Retrieve all the arch and distributions from the project
	prjArchDist, prjUUID, err := getDistAndArchFromPrj(prjIDToPass)
	if err != nil {
		return err
	}
	allArches := append(prjArchDist.MandArches, prjArchDist.OptArches...)
	allDistros := append(prjArchDist.MandDistros, prjArchDist.OptDistros...)

	for _, distro := range allDistros {
		for _, arch := range allArches {
			_, err = createBuildConfigFile(prjUUID, distro, arch, targetDir)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// createBuildConfigFile retrieves the build configuration file from
// the backend for one arch and one distro and write it in the given target directory
//
// :param prjUUID: UUID of the parent project
// :type prjUUID: uuid.UUID
// :param distro: Distribution wanted for the build config file
// :type distro: string
// :param arch: Architecture wanted for the build config file
// :type arch: string
// :param targetDir: Directory where the caller wants to create the file
// :type targetDir: string
// :returns: string, error -- the path to the file and an error if a problem occurred
func createBuildConfigFile(prjUUID uuid.UUID, distro, arch, targetDir string) (string, error) {
	// Retrieve the build configuration file from the project UUID
	configFile := rpapiv1.RPFile_Deprecated{}
	if err := bckAPI.ProjectGetBuildConfigFile(prjUUID.String(), distro, arch, &configFile); err != nil {
		return "", err
	}

	// Decode the base64 content
	fileContent, err := base64.StdEncoding.DecodeString(configFile.Content)
	if err != nil {
		return "", err
	}

	// Write the file in the right directory
	fullFilePath := path.Join(targetDir, configFile.Name)
	err = utils.WriteToFile(fullFilePath, string(fileContent))

	return fullFilePath, err
}

// getAllTargetTagsFromPrj retrieves all the target tags available
// for a project. The output is a map[distro]tag_released
//
// :param appParent: Parent of the application that the user wants to build
// :type appParent: rpapiv1.RPAppParent
// :returns: map[string]string, error -- map[distro]tag_released and an error if a problem occurred
func getAllTargetTagsFromPrj(appParent rpapiv1.RPAppParent) (map[string]string, error) {
	var err error
	result := make(map[string]string)
	var prjIDToPass string

	// If we do not have an UUID for the parent project, we need to retrieve it
	if utils.UUIDEqual(uuid.Nil, appParent.ProjectID) {
		prjIDToPass = appParent.ProjectSlug
	} else {
		prjIDToPass = appParent.ProjectID.String()
	}

	// Get the project
	project, err := getProjectFromIdentifier(prjIDToPass, true)
	if err != nil {
		return result, err
	}

	// Now, for every distribution inside the project, retrieve the corresponding distribution
	for _, currentDistro := range project.Distributions {
		if currentDistro.Set {
			// Get the target tag for the current distro
			tmpTargetTag, err := getTargetTagFromPrj(&project, currentDistro.Slug)
			if err != nil {
				return result, err
			}
			// Add the resulting target in the result map
			result[currentDistro.Slug] = tmpTargetTag
		}
	}

	return result, nil
}

// getTargetTagFromPrj retrieves the target tag wanted from a project pointer
// and a distro
//
// :param prj: Pointer to the project containing the "target tags" wanted
// :type prj: *rpapiv1.RPProjectFull
// :returns: string, error -- target tag and an error if a problem occurred
func getTargetTagFromPrj(prj *rpapiv1.RPProjectFull, distro string) (string, error) {
	targetTagResult := ""

	// Look inside the "repos" slice of the project where is the distribution that is interesting us
	for _, currentRepo := range prj.Repos {
		if currentRepo.Distro.Slug == distro {
			// We have found the right distro!

			// !!!!!!! FIXME !!!!!!!
			// Get the "release value" (now it is an URL). We will need to extract what we want, for example, in
			// http://iotbzh-app-test.lorient.iot/kbuild/repos-dist/ma-femme-s-appelle-maurice_1d4d7979--redpesk-devel-arz-release/latest/$arch
			// We want 'ma-femme-s-appelle-maurice_1d4d7979--redpesk-devel-arz-release'

			releaseURLTab := strings.Split(currentRepo.Release, "/")

			// Look for the only value that end with "-release"
			releaseValue := ""
			for _, text := range releaseURLTab {
				if strings.HasSuffix(text, "-release") {
					releaseValue = text
					break
				}
			}

			if releaseValue == "" {
				return targetTagResult, errors.New(noReleaseGotErr)
			}

			// Get rid of "-release" to have the target tag
			targetTagResult = strings.ReplaceAll(releaseValue, "-release", "")
			break
		}
	}

	// Check that a target tag has been found
	if targetTagResult == "" {
		errMsg := fmt.Sprintf(noDistroInProjectErr, distro, prj.Slug)
		return targetTagResult, errors.New(errMsg)
	}

	return targetTagResult, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------
