/*
Copyright © 2024 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"os"
	"path"
	"strings"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/ladapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"golang.org/x/term"
)

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initLocalNewCmd(rootCmd *cobra.Command) {

	// localCmd represents the local command
	var localNewCmd = &cobra.Command{
		Use:   localV2Command,
		Short: localV2CmdShortDesc,
		// Hidden: true,
	}

	// localExecCmd represents the "exec" command for local
	var localExecCmd = &cobra.Command{
		Use:          localExecCmdName,
		Short:        localExecCmdShortUsage,
		Long:         localExecCmdLongUsage,
		SilenceUsage: true,
		PreRunE:      initEventsRPLAD,
		RunE:         execCmdLine,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE", // No need for backend connection for this command
			"__lad_connect__": "TRUE"},
	}

	// localPkgCmd represents the "local package" command
	var localPkgCmd = &cobra.Command{
		Use:     localPkgCmdName,
		Aliases: []string{localPkgCmdAlias},
		Short:   localPkgCmdShortUsage,
	}

	// localPkgInitCmd represents the "local package init" command
	var localPkgInitCmd = &cobra.Command{
		Use:               localPkgInitCmdName,
		Short:             localPkgInitCmdShortUsage,
		Long:              localPkgInitCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              packageInitV2,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgFetchCmd represents the "local package fetch" command
	var localPkgFetchCmd = &cobra.Command{
		Use:               localV2FetchCmd,
		Short:             localV2FetchCmdShortDesc,
		Long:              localV2FetchCmdLongDesc,
		SilenceUsage:      true,
		RunE:              packageFetch,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
		},
	}

	// localPkgStatusCmd represents the "local package status" command
	var localPkgStatusCmd = &cobra.Command{
		Use:               localV2StatusCmd,
		Short:             localV2StatusCmdShortDesc,
		Long:              localV2StatusCmdLongDesc,
		SilenceUsage:      true,
		RunE:              packageStatus,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
		},
	}

	// localPkgCommitCmd represents the "local package commit" command
	var localPkgCommitCmd = &cobra.Command{
		Use:               localV2CommitCmd,
		Short:             localV2CommitCmdShortDesc,
		Long:              localV2CommitCmdLongDesc,
		SilenceUsage:      true,
		PreRunE:           initEventApp,
		RunE:              packageCommit,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// localPkgBuildCmd represents the "local package build" command
	var localPkgBuildCmd = &cobra.Command{
		Use:                localV2BuildCmd,
		Short:              localV2BuildCmdShortUsage,
		Long:               localV2BuildCmdLongUsage,
		SilenceUsage:       true,
		FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
		PreRunE:            initEventsRPLAD,
		RunE:               startBuildV2,
		ValidArgsFunction:  getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgServiceCmd represents the "local package service" command
	var localPkgServiceCmd = &cobra.Command{
		Use:                localV2ServiceCmd,
		Short:              localV2ServiceCmdShortDesc,
		SilenceUsage:       true,
		FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
		PreRunE:            initEventsRPLAD,
		RunE:               packageService,
		ValidArgsFunction:  getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgMetadataCmd represents the "local package metadata" command
	var localPkgMetadataCmd = &cobra.Command{
		Use:                localV2MetadataCmd,
		Short:              localV2MetadataCmdShortDesc,
		SilenceUsage:       true,
		FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
		PreRunE:            initEventsRPLAD,
		RunE:               packageMetadata,
		ValidArgsFunction:  getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgShellCmd represents the "local package shell" command
	var localPkgShellCmd = &cobra.Command{
		Use:               localV2ShellCmd,
		Short:             localV2ShellCmdShortUsage,
		Long:              localV2ShellCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              startShellV2,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgCleanCmd represents the "local package clean" command
	var localPkgCleanCmd = &cobra.Command{
		Use:               localPkgCleanCmdName,
		Short:             localPkgCleanCmdShortUsage,
		Long:              localPkgCleanCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              packageClean,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__lad_connect__":          "TRUE"},
	}

	// localCheckSetUpCmd represents the "local check-setup" command
	var localCheckSetUpCmd = &cobra.Command{
		Use:          localV2CheckSetupCmd,
		Short:        localV2CheckSetupCmdShortUsage,
		SilenceUsage: true,
		PreRunE:      initEventsRPLAD,
		RunE:         localCheckSetup,
		Annotations:  map[string]string{"__lad_connect__": "TRUE"},
	}

	// connectionsCmd represents the "local connections" command
	var connectionsCmd = &cobra.Command{
		Use:   connectionsCommand,
		Short: localConnectionsCmdShortDesc,
	}

	// connectionsListCmd represents the "local connections list" command
	var connectionsListCmd = &cobra.Command{
		Use:         connectionsListCmd,
		Short:       fmt.Sprintf(connectionsListCmdShortDesc, "local-builders"),
		Long:        fmt.Sprintf(connectionsListCmdLongDesc, "local-builders"),
		RunE:        listConnections,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// getCmd represents the get command
	var connectionsGetCmd = &cobra.Command{
		Use:               connectionsGetCmd,
		Short:             connectionsGetCmdShortUsage,
		RunE:              getConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// connectionsCurrentCmd represents the current command
	var connectionsCurrentCmd = &cobra.Command{
		Use:         connectionsCurrentCmd,
		Short:       connectionsCurrentCmdShortUsage,
		RunE:        currentConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// addCmd represents the get command
	var connectionsAddCmd = &cobra.Command{
		Use:         connectionsAddCmd,
		Short:       connectionsAddShortDesc,
		Long:        fmt.Sprintf(connectionsAddLongDesc, "local-builders"),
		RunE:        addConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// updateCmd represents the update command
	var connectionsUpdateCmd = &cobra.Command{
		Use:               connectionsUpdateCmd,
		Short:             connectionsUpdateShortDesc,
		RunE:              updateConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// deleteCmd represents the delete command
	var connectionsDeleteCmd = &cobra.Command{
		Use:               connectionsDeleteCmd,
		Short:             fmt.Sprintf(connectionsDeleteCmdShortUsage, "local-builder"),
		RunE:              deleteConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// useCmd represents the use command
	var connectionsUseCmd = &cobra.Command{
		Use:               connectionsUseCmd,
		Short:             fmt.Sprintf(connectionsUseCmdShortDesc, "local-builder"),
		RunE:              useConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	localPkgCmd.PersistentFlags().StringP(localPkgCmdSourceFlag, localPkgCmdSourceFlagAlias, "./", localPkgCmdSourceFlagUsage)
	localPkgCmd.PersistentFlags().StringP(localPkgCmdPackageFlag, localPkgCmdPackageFlagAlias, "./../gitpkgs/", localPkgCmdPackageFlagUsage)
	// ----------------------------------------------------------

	// Bind the cobra flags to viper
	viper.BindPFlag(localPkgCmdSourceFlag, localPkgCmd.PersistentFlags().Lookup(localPkgCmdSourceFlag))   // Corresponding environment var: RP_SOURCESDIR
	viper.BindPFlag(localPkgCmdPackageFlag, localPkgCmd.PersistentFlags().Lookup(localPkgCmdPackageFlag)) // Corresponding environment var: RP_PKGSDIR
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Exec
	localExecCmd.Flags().StringP(localExecCmdWorkingDirFlag, localExecCmdWorkingDirFlagAlias, "", localExecCmdWorkingDirFlagUsage)
	// Package init
	localPkgInitCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	localPkgInitCmd.Flags().StringP(localV2PkgInitBranchFlag, localV2PkgInitBranchFlagAlias, "", localV2PkgInitBranchFlagUsage)
	localPkgInitCmd.Flags().BoolP(localV2PkgInitForceFlag, localV2PkgInitForceFlagAlias, false, localV2PkgInitForceFlagUsage)
	// Package fetch
	localPkgFetchCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	localPkgFetchCmd.Flags().BoolP(localV2FetchCmdGitpkgFlag, localV2FetchCmdGitpkgFlagAlias, false, localV2FetchCmdGitpkgFlagUsage)
	localPkgFetchCmd.Flags().BoolP(localV2FetchCmdMockCfgFlag, localV2FetchCmdMockCfgFlagAlias, false, localV2FetchCmdMockCfgFlagUsage)
	localPkgFetchCmd.Flags().StringP(localV2PkgInitBranchFlag, localV2PkgInitBranchFlagAlias, "", localV2PkgInitBranchFlagUsage)
	localPkgFetchCmd.Flags().BoolP(localV2FetchCmdForceFlag, localV2FetchCmdForceFlagAlias, false, localV2FetchCmdForceFlagUsage)
	// Package status
	localPkgStatusCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	// Package commit
	localPkgCommitCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	// Package build
	localPkgBuildCmd.Flags().StringP(localV2BuildProjectFlag, localV2BuildProjectFlagAlias, "", localV2BuildProjectFlagUsage)
	localPkgBuildCmd.Flags().StringP(localV2BuildArchFlag, localV2BuildArchFlagAlias, localV2BuildArchFlagDefault, localV2BuildArchFlagUsage)
	localPkgBuildCmd.Flags().StringP(localV2BuildDistroFlag, localV2BuildDistroFlagAlias, "", localV2BuildDistroFlagUsage)
	// Package service
	localPkgServiceCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	// Package metadata
	localPkgMetadataCmd.Flags().String(localPkgInitProjectFlag, "", localPkgInitProjectFlagUsage)
	// Package Shell
	localPkgShellCmd.Flags().StringP(localV2BuildProjectFlag, localV2BuildProjectFlagAlias, "", localV2BuildProjectFlagUsage)
	localPkgShellCmd.Flags().StringP(localV2BuildArchFlag, localV2BuildArchFlagAlias, localV2BuildArchFlagDefault, localV2BuildArchFlagUsage)
	localPkgShellCmd.Flags().StringP(localV2BuildDistroFlag, localV2BuildDistroFlagAlias, "", localV2BuildDistroFlagUsage)
	// Package clean
	localPkgCleanCmd.Flags().BoolP(localPkgCleanInitFlag, localPkgCleanInitFlagAlias, false, localPkgCleanInitFlagUsage)
	localPkgCleanCmd.Flags().String(localPkgCleanProjectFlag, "", localPkgCleanProjectFlagUsage)
	localPkgCleanCmd.Flags().BoolP(localPkgCleanForceFlag, localPkgCleanForceFlagAlias, false, localPkgCleanForceFlagUsage)
	localPkgCleanCmd.Flags().BoolP(localV2CleanNoMockFlag, localV2CleanNoMockFlagAlias, false, localV2CleanNoMockFlagUsage)
	// Connections list
	connectionsListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections get
	connectionsGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections current
	connectionsCurrentCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections add
	connectionsAddCmd.Flags().StringP(connectionsAddIPFlag, connectionsAddIPFlagAlias, "", connectionsAddIPFlagUsage)
	// Connections update
	connectionsUpdateCmd.Flags().StringP(connectionsUpdateIPFlag, connectionsUpdateIPFlagAlias, flagUpdateKeepCurrentValue, connectionsUpdateIPFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required

	// Add some flags as hidden

	// Specify dynamic flag completion
	localPkgInitCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgFetchCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgStatusCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgCommitCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgBuildCmd.RegisterFlagCompletionFunc(localPkgRunProjectFlag, getListPrjsAsTab)
	localPkgBuildCmd.RegisterFlagCompletionFunc(localV2BuildArchFlag, getListArchAsTab)
	localPkgBuildCmd.RegisterFlagCompletionFunc(localV2BuildDistroFlag, getListDistAsTab)
	localPkgServiceCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgMetadataCmd.RegisterFlagCompletionFunc(localPkgInitProjectFlag, getListPrjsAsTab)
	localPkgShellCmd.RegisterFlagCompletionFunc(localV2BuildArchFlag, getListArchAsTab)
	localPkgShellCmd.RegisterFlagCompletionFunc(localV2BuildDistroFlag, getListDistAsTab)
	localPkgCleanCmd.RegisterFlagCompletionFunc(localPkgCleanProjectFlag, getListPrjsFromBldDirAsTab)

	// Add the command "local-V2" as a subcommand of rp-cli
	rootCmd.AddCommand(localNewCmd)
	// Add the command "exec" as a subcommand of localCmd
	localNewCmd.AddCommand(localExecCmd)
	// Add the command "package" as a subcommand of localCmd
	localNewCmd.AddCommand(localPkgCmd)
	// Add the command "connections" as a subcommand of localCmd
	localNewCmd.AddCommand(connectionsCmd)
	// Add the command "init" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgInitCmd)
	// Add the command "fetch" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgFetchCmd)
	// Add the command "status" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgStatusCmd)
	// Add the command "commit" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgCommitCmd)
	// Add the command "build" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgBuildCmd)
	// Add the command "service" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgServiceCmd)
	// Add the command "metadata" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgMetadataCmd)
	// Add the command "shell" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgShellCmd)
	// Add the command "clean" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgCleanCmd)
	// Add the command "check-setup" as a subcommand of localCmd
	localNewCmd.AddCommand(localCheckSetUpCmd)
	// Add the command "list" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsListCmd)
	// Add the command "get" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsGetCmd)
	// Add the command "current" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsCurrentCmd)
	// Add the command "add" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsAddCmd)
	// Add the command "update" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsUpdateCmd)
	// Add the command "delete" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsDeleteCmd)
	// Add the command "use" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsUseCmd)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// packageInit initializes everything that is needed to package
// an application inside the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageInitV2(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine0)

	// Before anything check that 'git' and 'git-lfs' are installed in the system
	_, err := utils.RunCmdInDirectory("/tmp/", "git", "--version")
	if err != nil {
		return errors.New(gitPkgNotInstalled)
	}

	// Get the "force" flag value
	forceFlag, err := cmd.Flags().GetBool(localV2PkgInitForceFlag)
	if err != nil {
		logFlagError(localV2PkgInitForceFlag)
		return err
	}

	// Get the first main infos we need
	fullConfiguration, appStructure, err := prePackagingRoutineV2(cmd, args)
	if err != nil {
		return err
	}

	// Display something like "Retrieving packaging configurations" here
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine5, appStructure.Slug)

	// Get the branch flag
	fullConfiguration.Branch, err = cmd.Flags().GetString(localV2PkgInitBranchFlag)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something like "Setting up the local-builder" here
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine6)

	// Set-up the local builder
	err = localBuilderSetUpV2(&appStructure, &fullConfiguration, forceFlag)
	if err != nil {
		return err
	}

	return nil
}

// packageStatus shows the status of files in a local working directory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageStatus(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localV2StatusCmdAnsLine0)

	// Get full context -----------------------------------
	appStructure, _, rootGitpkgDir, err := getLocalV2Context(cmd, args, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	fullGitpkgDir := path.Join(rootGitpkgDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Get the 'local_build_config.json' file to print the branch used
	localCfg, _, err := getLocalBuildConfig(fullGitpkgDir, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Print the branch used
	if localCfg.Branch != "" {
		RPPrinter.CustomPrintLn()
		RPPrinter.CustomPrintF(localV2StatusCmdAnsLine2, localCfg.Branch)
	}

	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintLn(localV2StatusCmdAnsLine1)

	// Finally, check the status
	_, _, err = checkGitpkgsStatus(fullGitpkgDir, true, false)

	RPPrinter.CustomPrintLn()

	return err

}

// packageFetch fetches the app content and the local-builder configuration
// in order to be able to build it locally
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageFetch(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localV2FetchCmdAnsLine0)

	// Get the value of the different flags ---------------
	gitpkgFlag, err := cmd.Flags().GetBool(localV2FetchCmdGitpkgFlag)
	if err != nil {
		logFlagError(localV2FetchCmdGitpkgFlag)
		return err
	}
	localBuilderFlag, err := cmd.Flags().GetBool(localV2FetchCmdMockCfgFlag)
	if err != nil {
		logFlagError(localV2FetchCmdMockCfgFlag)
		return err
	}
	forceFlag, err := cmd.Flags().GetBool(localV2FetchCmdForceFlag)
	if err != nil {
		logFlagError(localV2FetchCmdForceFlag)
		return err
	}
	getEverything := !gitpkgFlag && !localBuilderFlag

	// Get the branch flag
	branch, err := cmd.Flags().GetString(localV2PkgInitBranchFlag)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get full context -----------------------------------
	appStructure, rootSrcDir, rootGitpkgDir, err := getLocalV2Context(cmd, args, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	localConfig := RPLocalBuildConfigV2{
		SourcesDir:  rootSrcDir,
		PackagesDir: rootGitpkgDir,
		Branch:      branch,
	}

	// Now do the work!
	return fetchGitpkgAndMockFiles(&localConfig,
		&appStructure,
		(gitpkgFlag || getEverything),
		(localBuilderFlag || getEverything),
		forceFlag)
}

// packageCommit uploads files that have been modified/added/deleted in
// the working directory, to the remote gitpkg in the redpesk factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageCommit(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localV2CommitCmdAnsLine0)

	// Get full context -----------------------------------
	appStructure, _, rootGitpkgDir, err := getLocalV2Context(cmd, args, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	fullGitpkgDir := path.Join(rootGitpkgDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintLn(localV2StatusCmdAnsLine1)

	// Get status details ---------------------------------
	statusBool, modifDetails, err := checkGitpkgsStatus(fullGitpkgDir, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------
	RPPrinter.CustomPrintLn()

	// If the status is true, we have nothing to commit!
	if statusBool {
		RPPrinter.CustomPrintLn(localV2CommitNothingToCommit)
		return nil
	}

	// Get the list of files that the user may have given to rp-cli
	userFiles := []string{}
	if len(args) > 1 {
		userFiles = args[1:]
	}

	// If we reached here, this means that we have some files to upload/remove

	// Get the branch used to fetch the gitpkgs files
	branchConfiguration, _, err := getLocalBuildConfig(fullGitpkgDir, false)
	if err != nil {
		return err
	}

	// Get the active branch to see if we need to change the branch
	activeBranch, err := getActiveBranch(appStructure.ID)
	if err != nil {
		return err
	}

	// Switch to specified branch if needed
	if activeBranch != branchConfiguration.Branch {
		err = switchToExistingBranch(appStructure.ID, branchConfiguration.Branch)
		if err != nil {
			return err
		}

		// Defer switching back to current branch
		defer switchToExistingBranch(appStructure.ID, activeBranch)
	}

	// Upload all added/modified files
	filesToUpload := append(modifDetails.Modified, modifDetails.Added...)
	for _, tmpFile := range filesToUpload {

		// Filter the files to upload
		if (len(userFiles) > 0) && !utils.ContainsString(userFiles, tmpFile) {
			continue
		}

		RPPrinter.CustomPrintF(localV2CommitUpload, tmpFile)

		// Now upload the current file
		fullPathFile := path.Join(fullGitpkgDir, tmpFile)
		err = uploadFileUtility(fullPathFile, appStructure.ID, true)
		if err != nil {
			return err
		}
	}

	// Remove all removed files
	for _, tmpFile := range modifDetails.Deleted {

		// Filter the files to delete
		if (len(userFiles) > 0) && !utils.ContainsString(userFiles, tmpFile) {
			continue
		}

		RPPrinter.CustomPrintF(localV2CommitRemoval, tmpFile)

		// Now remove the current file
		err := bckAPI.ApplicationRemoveFile(appStructure.ID.String(), tmpFile)
		if err != nil {
			Log.Error("error when removing ", tmpFile, ": ", err)
			return err
		}
	}

	// Update the md5sum file
	return updateOrCreateMD5SumFile(fullGitpkgDir, userFiles)
}

// startBuildV2 builds locally the application passed as argument
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func startBuildV2(cmd *cobra.Command, args []string) error {
	// Before anything, handle the case where the user is asking for help
	helpAsked, err := handleIotpkgSubCmdHelp(cmd, originalArgs, "build")
	if err != nil {
		return err
	}
	if helpAsked {
		// Reset the global variable (mainly for tests)
		originalArgs = []string{}
		return nil
	}

	RPPrinter.CustomPrintLn(localV2BuildRunCmdAnsLine0)

	// Parse rp-cli flags and extract iotpkg args ---------
	iotpkgArgs, err := getIotpkgArgs(cmd)
	if err != nil {
		return err
	}
	Log.Debugf("[LOCAL-V2] Iotpkg args: %+v\n", iotpkgArgs)
	// ----------------------------------------------------

	// Get local context -----------------------------------
	appStructure, _, fullGitPkgDir, err := getLocalV2Context(cmd, cmd.Flags().Args(), true)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get the build context (arch and distro) ------------
	archToBld, distroToBld, err := getLocalBuildContext(cmd, &appStructure)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Create the gitpkg directory path from the local-builder POV
	localBldGitpkgDir := path.Join(localBuilderGitPkgsDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Add the flag "--repofile" if it is not already set
	if !utils.ContainsString(iotpkgArgs, "--repofile") && !utils.ContainsString(iotpkgArgs, "-r") {
		iotpkgArgs = append(iotpkgArgs, "--repofile", path.Join(localBldGitpkgDir, localCfgSubDirectory, appStructure.Parent.ProjectSlug+".json"))
	}

	// Add the distro and the architecture flag
	iotpkgArgs = append(iotpkgArgs, "--arch", archToBld, "--distro", distroToBld)

	buildRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBldGitpkgDir,
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgRPBuildPost(&buildRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)

	// Construct the directory containing the results
	resultsPath := fmt.Sprintf("%s/results_%s/", fullGitPkgDir, appStructure.Slug)

	// Display the info to the user
	RPPrinter.CustomColorPrintLn(printer.ColorCyan, localPkgRunCmdAnsLine4, resultsPath)
	RPPrinter.CustomPrintLn()

	// Congrats the user if the packaging went well!
	if err == nil {
		RPPrinter.CustomColorPrintLn(printer.ColorGreen, localV2BuildRunCmdAnsResultGood)
	} else {
		RPPrinter.CustomColorPrintLn(printer.ColorRed, localV2BuildRunCmdAnsResultBad)
	}

	return err
}

// startBuildV2 builds locally the application passed as argument
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func startShellV2(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localV2ShellCmdAnsLine0)

	// Get local context -----------------------------------
	appStructure, _, _, err := getLocalV2Context(cmd, cmd.Flags().Args(), true)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get the build context (arch and distro) ------------
	archToBld, distroToBld, err := getLocalBuildContext(cmd, &appStructure)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Create the gitpkg directory path from the local-builder POV
	localBldGitpkgDir := path.Join(localBuilderGitPkgsDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Add the flag "--repofile" in the flags, also the distro, the architecture flag and finally, the shell flag
	iotpkgArgs := []string{
		"--repofile", path.Join(localBldGitpkgDir, localCfgSubDirectory, appStructure.Parent.ProjectSlug+".json"),
		"--arch", archToBld,
		"--distro", distroToBld,
		"--shell",
	}

	buildRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBldGitpkgDir,
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgRPBuildPost(&buildRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)

	return err
}

// packageService allow to set app service value using `iotpkg` inside
// the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageService(cmd *cobra.Command, args []string) error {

	// Before anything, handle the case where the user is asking for help
	helpAsked, err := handleIotpkgSubCmdHelp(cmd, originalArgs, "service")
	if err != nil {
		return err
	}
	if helpAsked {
		// Reset the global variable (mainly for tests)
		originalArgs = []string{}
		return nil
	}

	RPPrinter.CustomPrintLn(localV2ServiceCmdAnsLine0)

	// Parse rp-cli flags and extract iotpkg args ---------
	iotpkgArgs, err := getIotpkgArgs(cmd)
	if err != nil {
		return err
	}
	Log.Debugf("[LOCAL-V2] Iotpkg args: %+v\n", iotpkgArgs)
	// ----------------------------------------------------

	// Get full context -----------------------------------
	appStructure, _, _, err := getLocalV2Context(cmd, cmd.Flags().Args(), true)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	serviceRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: path.Join(localBuilderGitPkgsDir, appStructure.Parent.ProjectSlug, appStructure.PackageName),
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgServicePost(&serviceRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)

	return err
}

// packageMetadata allow to set app metadata value using `iotpkg` inside
// the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageMetadata(cmd *cobra.Command, args []string) error {

	// Before anything, handle the case where the user is asking for help
	helpAsked, err := handleIotpkgSubCmdHelp(cmd, originalArgs, "metadata")
	if err != nil {
		return err
	}
	if helpAsked {
		return nil
	}

	RPPrinter.CustomPrintLn(localV2MetadataCmdAnsLine0)

	// Parse rp-cli flags and extract iotpkg args ---------
	iotpkgArgs, err := getIotpkgArgs(cmd)
	if err != nil {
		return err
	}
	Log.Debugf("[LOCAL-V2] Iotpkg args: %+v\n", iotpkgArgs)
	// ----------------------------------------------------

	// Get full context -----------------------------------
	appStructure, _, _, err := getLocalV2Context(cmd, cmd.Flags().Args(), true)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	metadataRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: path.Join(localBuilderGitPkgsDir, appStructure.Parent.ProjectSlug, appStructure.PackageName),
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgMetadataPost(&metadataRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)

	return err
}

// localCheckSetup check that everything needed to local build
// is correctly set-up on the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func localCheckSetup(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localV2CheckSetupCmdAnsLine0)

	defer RPPrinter.ForceFlush()

	// Check factory connection ---------------------------
	RPPrinter.CustomPrint(localV2CheckSetupCmdAnsLine1)
	cfg := rpapiv1.RPAPIConfig{}
	if err := bckAPI.RPConfigGet(&cfg); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check local-builder connection ---------------------
	RPPrinter.CustomPrint(localV2CheckSetupCmdAnsLine2)
	_, err := localBAPI.GetVersion()
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check mount-points ---------------------------------
	RPPrinter.CustomPrint(localV2CheckSetupCmdAnsLine3)
	err = checkLocalBuilderMountPoints()
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check QEMU set-up ----------------------------------
	RPPrinter.CustomPrint(localV2CheckSetupCmdAnsLine4)
	err = checkQemuSetUp()
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// createLocalBuilderWorkingDirs is the function that manage the creation of the directories
// used to build inside the local-builder
//
// :param rootSrcDir: Root directory containing the source directories
// :type rootSrcDir: string
// :param rootGitpkgDir: Root directory containing the gitpkg directories
// :type rootGitpkgDir: string
// :param prjSlug: Project slug
// :type prjSlug: string
// :param appPkgName: Application package name
// :type appPkgName: string
// :returns: error -- an error if something went wrong
func createLocalBuilderWorkingDirs(rootSrcDir, rootGitpkgDir, prjSlug, appPkgName string) error {

	// Create the path to the directories that are interesting us ------------------
	// Project directories
	ourSourcesPrjDir := path.Join(rootSrcDir, prjSlug)
	ourPackagesPrjDir := path.Join(rootGitpkgDir, prjSlug)
	// App directories
	ourSourcesDir := path.Join(ourSourcesPrjDir, appPkgName)
	ourPackagesDir := path.Join(ourPackagesPrjDir, appPkgName)
	// rp-cli configuration directories
	ourCfgPkgDir := path.Join(ourPackagesDir, localCfgSubDirectory)
	// -----------------------------------------------------------------------------

	// Create the project level directories if there are not existing
	err := utils.CreateDirIfNotExists(ourSourcesPrjDir, false)
	if err != nil {
		return err
	}
	err = utils.CreateDirIfNotExists(ourPackagesPrjDir, false)
	if err != nil {
		return err
	}

	// Now create the app slug sub-directories
	err = utils.CreateDirIfNotExists(ourSourcesDir, false)
	if err != nil {
		return err
	}
	err = utils.CreateDirIfNotExists(ourPackagesDir, false)
	if err != nil {
		return err
	}

	// Finally the configuration sub-directory
	err = utils.CreateDirIfNotExists(ourCfgPkgDir, false)
	if err != nil {
		return err
	}

	return nil
}

// checkGitpkgsStatus is the function that check if some modification have been done on the
// gitpkgs files
//
// :param fullGitpkgDir: Full path to the directory containing the files to check
// :type fullGitpkgDir: string
// :param displayRes: Ask to display details of difference
// :type displayRes: bool
// :param quiet: If true, the function does not return an error if the md5sum file does not exist
// :type quiet: bool
// :returns: bool, RPLocalGitpkgsModification, error -- true if the status is OK, false in the other case, the details of modification and an error if something went wrong
func checkGitpkgsStatus(fullGitpkgDir string, displayRes, quiet bool) (bool, RPLocalGitpkgsModification, error) {
	status := true
	statusDetails := RPLocalGitpkgsModification{
		Added:    []string{},
		Modified: []string{},
		Deleted:  []string{},
	}

	// Firstly check if the md5sum file exist
	md5sumFile := path.Join(fullGitpkgDir, localCfgSubDirectory, localMD5SUMFile)

	// Return an error if the md5sum file does not exist
	_, err := os.Stat(md5sumFile)
	if os.IsNotExist(err) && !quiet {
		return false, statusDetails, errors.New(md5sumFileDoesNotExist)
	} else if os.IsNotExist(err) && quiet {
		// Return true if the function is quiet and that the md5sum does not exist
		return true, statusDetails, nil
	}

	// Read the JSON file
	md5sumMap := make(map[string]string)
	err = utils.ReadJsonFile(md5sumFile, &md5sumMap)
	if err != nil {
		return false, statusDetails, err
	}

	// Get all the file currently in the gitpkg "repo"
	currentLocalFiles, err := getLocalGitpkgsFileOrDirsList(fullGitpkgDir, true)
	if err != nil {
		return false, statusDetails, err
	}
	filesList := currentLocalFiles["files"]

	// Now for each file, compare calculate the md5sum and compare with what we got in the file
	for filename, filemd5sum := range md5sumMap {

		// Create the path for the current file
		currFilePath := path.Join(fullGitpkgDir, filename)

		// Delete the file from our "current local" list
		utils.DeleteStringFromSlice(&filesList, currFilePath)

		// Check that the file still exist
		_, err = os.Stat(currFilePath)
		if os.IsNotExist(err) {
			// The file does not exist anymore
			status = false
			statusDetails.Deleted = append(statusDetails.Deleted, filename)
			if displayRes {
				RPPrinter.CustomColorPrintF(printer.ColorRed, localV2StatusDelFileLine, filename)
			}
			continue
		}

		// If the file exists, check the md5sum
		localMd5sum, err := utils.CalculateMd5SumOnFile(currFilePath, true)
		if err != nil {
			return false, statusDetails, err
		}
		if localMd5sum != filemd5sum {
			// The file has been modified
			status = false
			statusDetails.Modified = append(statusDetails.Modified, filename)
			if displayRes {
				RPPrinter.CustomColorPrintF(printer.ColorRed, localV2StatusModifiedFileLine, filename)
			}
		}
	}

	// Create the directory prefix to delete from file path
	dirPrefix := fullGitpkgDir
	if !strings.HasSuffix(dirPrefix, "/") {
		dirPrefix += "/"
	}

	// Now add every file that stayed in our "current local" list, as an "added" file
	for _, tmpAddFile := range filesList {
		status = false
		// Change the file path to convert to the local gitpkg path
		addFile := strings.ReplaceAll(tmpAddFile, dirPrefix, "")
		// Add it to the "added" files list
		statusDetails.Added = append(statusDetails.Added, addFile)
		if displayRes {
			RPPrinter.CustomColorPrintF(printer.ColorRed, localV2StatusAddedFileLine, addFile)
		}
	}

	if status && displayRes {
		RPPrinter.CustomColorPrintLn(printer.ColorGreen, localV2StatusNoChanges)
	}

	return status, statusDetails, nil
}

// getLocalGitpkgsFileOrDirsList retrieve the list of all the files/dirs currently present in the gitpkg directory given.
// It filters out the files that can be found in the gitpkg directory but that are not part of gitpkg (like ./rp-cli content)
//
// :param fullGitpkgDir: Full path to the directory containing the gitpkg files
// :type fullGitpkgDir: string
// :param filesOnly: Tell the function to not list the directories
// :type filesOnly: bool
// :returns: map[string][]string error -- lists of files and subdirectories and an error if something went wrong
func getLocalGitpkgsFileOrDirsList(fullGitpkgDir string, filesOnly bool) (map[string][]string, error) {
	// Get the list of all files in the gitpkg directory
	tmpList, err := utils.FindAllFilesAndDirsRecursively(fullGitpkgDir)
	if err != nil {
		return tmpList, err
	}

	// Now filter out the .rp-cli/ lines
	actualMap := make(map[string][]string)
	if !filesOnly {
		actualMap["dirs"] = []string{}
		for _, tmpDirName := range tmpList["dirs"] {
			if !strings.Contains(tmpDirName, localCfgSubDirectory) && (tmpDirName != fullGitpkgDir) {
				actualMap["dirs"] = append(actualMap["dirs"], tmpDirName)
			}
		}
	}
	actualMap["files"] = []string{}
	for _, tmpFileName := range tmpList["files"] {
		if !strings.Contains(tmpFileName, localCfgSubDirectory) {
			actualMap["files"] = append(actualMap["files"], tmpFileName)
		}
	}

	return actualMap, nil
}

// getLocalBuildConfig retrieves the local-builder configuration in the specified path.
// The function adds the .rp-cli/ suffix if it is not included.
// Additionally, it can output an error on the "local_build_config.json" file not existing if asked
//
// :param packageDir:Path to the directory containing the configuration file (with or without '.rp-cli')
// :type packageDir: string
// :param errIfNotExist: If set, the function returns an error when "local_build_config.json" does not exist
// :type errIfNotExist: bool
// :returns: RPLocalBuildConfig, string error -- the local-builder configuration, the path to the config file and an error if something went wrong
func getLocalBuildConfig(packageDir string, errIfNotExist bool) (RPLocalBuildConfig, string, error) {
	// Create the return value
	localConfig := RPLocalBuildConfig{}

	// Work on the package directory given
	fullCfgPath := packageDir
	if !strings.Contains(".rp-cli", path.Base(packageDir)) {
		fullCfgPath = path.Join(packageDir, localCfgSubDirectory)
	}

	// Create the config file path
	localBldConfFile := path.Join(fullCfgPath, "local_build_config.json")

	// Check if file exists
	_, err := os.Stat(localBldConfFile)
	if err != nil && !os.IsNotExist(err) {
		return localConfig, localBldConfFile, err
	} else if err == nil {
		Log.Debugln("[FETCH] Retrieving local config from:", localBldConfFile)
		err = utils.ReadJsonFile(localBldConfFile, &localConfig)
		if err != nil {
			return localConfig, localBldConfFile, err
		}
	} else if os.IsNotExist(err) {
		Log.Debugln("[FETCH] 'local_build_config.json' does not exist, nothing to retrieve")
		if errIfNotExist {
			return localConfig, localBldConfFile, err
		}
	}
	return localConfig, localBldConfFile, nil
}

// updateOrCreateMD5SumFile is the function that takes as input the gitpkg directory and creates or update a
// json file containing all the md5sum of the files in the .rp-cli/ subdirectory
//
// :param fullGitpkgDir: Full path to the directory containing the files to check
// :type fullGitpkgDir: string
// :param fileNamesToUpdate: List of files to update in the file (if empty, the file is just re-created)
// :type fileNamesToUpdate: []string
// :returns: error -- an error if something went wrong
func updateOrCreateMD5SumFile(fullGitpkgDir string, fileNamesToUpdate []string) error {

	// Create the object that will store the previous values if necessary
	oldMd5sumMap := make(map[string]string)
	// Path to the md5sum file
	md5sumFile := path.Join(fullGitpkgDir, localCfgSubDirectory, localMD5SUMFile)

	// Check if the file exists (only if the user gave some names, in the other case -> full update)
	if _, err := os.Stat(md5sumFile); err == nil && len(fileNamesToUpdate) > 0 {
		// If the file exists, get the params already existing inside
		err := utils.ReadJsonFile(md5sumFile, &oldMd5sumMap)
		if err != nil {
			return err
		}
	}

	// Get the list of files to check
	objectsInDir, err := getLocalGitpkgsFileOrDirsList(fullGitpkgDir, true)
	if err != nil {
		return err
	}

	// Create the directory prefix
	dirPrefix := fullGitpkgDir
	if !strings.HasSuffix(dirPrefix, "/") {
		dirPrefix += "/"
	}

	// Rework the filenames to update to add the directory prefix
	fullFileNamesToUpdate := []string{}
	for _, fileToUpdate := range fileNamesToUpdate {
		fullFileName := path.Join(dirPrefix, fileToUpdate)
		fullFileNamesToUpdate = append(fullFileNamesToUpdate, fullFileName)

		// Check if the filename is in the files list found on the disk
		// If it is not the case, this means that the modification on the file is a deletion
		if !utils.ContainsString(objectsInDir["files"], fullFileName) {
			// Delete from oldMd5sumMap
			delete(oldMd5sumMap, fileToUpdate)
		}
	}

	// Iterate over files to calculate their md5sum
	for _, file := range objectsInDir["files"] {
		// Filter the files to update if the user gave some
		if utils.ContainsString(fullFileNamesToUpdate, file) || (len(fullFileNamesToUpdate) == 0) {
			tmpMd5sum, err := utils.CalculateMd5SumOnFile(file, true)
			if err != nil {
				return err
			}
			// Get rid of the directory prefix and save the md5sum in the map
			tmpFileName := strings.ReplaceAll(file, dirPrefix, "")
			// Save in map
			oldMd5sumMap[tmpFileName] = tmpMd5sum
		}
	}

	// Write in md5sum file
	return utils.WriteJsonFile(md5sumFile, oldMd5sumMap)
}

// localBuilderSetUpV2 fills the RPLocalBuildConfig structure with the values left
// to set-up fully the local-builder to build
//
// :param app: Structure containing all needed data for application
// :type app: *rpapiv1.RPApplicationFull
// :param config: Configuration to set for the local builder
// :type config: *RPLocalBuildConfigV2
// :param overwrite: Overwrite any existing directory
// :type overwrite: bool
// :returns: error -- an error if something went wrong
func localBuilderSetUpV2(app *rpapiv1.RPApplicationFull, localConfig *RPLocalBuildConfigV2, overwrite bool) error {

	// --------------------------------------------------------------
	// -----             '<app_gitpkgs>/' directory             -----
	// --------------------------------------------------------------
	err := fetchGitpkgAndMockFiles(localConfig, app, true, true, overwrite)
	if err != nil {
		return err
	}

	// --------------------------------------------------------------
	// -----           '<app_gitsources>/' directory            -----
	// --------------------------------------------------------------
	// Clone the sources inside gitsources/
	err = cloneSourcesRepo(app, localConfig, overwrite)
	if err != nil {
		return err
	}

	return nil
}

// getLocalAppV2 parses the user command line and retrieve the
// application UUID and slug. It also permits to check that the corresponding
// project is created and if the application exists in Redpesk
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: rpapiv1.RPApplicationFull, error -- the application and an error if something went wrong
func getLocalAppV2(cmd *cobra.Command, args []string) (rpapiv1.RPApplicationFull, error) {
	// Check if the project and the application exists, and get it
	return genericGetAppFromArgs(cmd, args, true)
}

// createRepoInheritanceFile retrieves the repo inheritance file from
// the backend and write it in the given target directory
//
// :param appParent: Parent of the application that the user wants to build
// :type appParent: rpapiv1.RPAppParent
// :param targetDir: Directory where the caller wants to create the file
// :type targetDir: string
// :returns: error -- an error if a problem occurred
func createRepoInheritanceFile(appParent rpapiv1.RPAppParent, targetDir string) error {
	var err error

	repoInheritanceFile := path.Join(targetDir, (appParent.ProjectSlug + ".json"))

	// Get the connection parameters
	serverAlias := viper.GetString(serverAliasFlag)
	connecParams, err := getConnectionParams(serverAlias, accessTokenFile)
	if err != nil {
		return err
	}

	return bckAPI.ProjectGetReposInheritance(appParent.ProjectID.String(), repoInheritanceFile, connecParams.ServerURL)
}

// getLocalV2Context retrieves the structures necessary for several local-v2 subcommands
// These include:
//   - The full application structure
//   - Path to the main source directory (or full path if getFullDir is True)
//   - Path to the main gitpkg directory (or full path if getFullDir is True)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param getFullDir: Ask the function to get and check the full directories path and not just the root
// :type getFullDir: bool
// :returns: rpapiv1.RPApplicationFull, string, string, error -- app, rootSrcDir, rootGitpkgDir and an error if a problem occurred
func getLocalV2Context(cmd *cobra.Command, args []string, getFullDir bool) (rpapiv1.RPApplicationFull, string, string, error) {
	// Get applications details ---------------------------
	// Display info
	RPPrinter.CustomPrint(localV2FetchCmdAnsLine1)
	// Now, retrieve the app ID
	appStructure, err := getLocalAppV2(cmd, args)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return rpapiv1.RPApplicationFull{}, "", "", err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Fill a directory suffix if the caller wants to check the full directory
	dirSuffix := ""
	if getFullDir {
		dirSuffix = path.Join(appStructure.Parent.ProjectSlug, appStructure.PackageName)
	}

	// Check the directories to use for the local-builder -
	// Display info
	RPPrinter.CustomPrint(localV2FetchCmdAnsLine2)
	// Check that the local builder directories exist
	rootSrcDir, rootGitpkgDir, err := getAndChecksBuildDirs(dirSuffix)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return rpapiv1.RPApplicationFull{}, "", "", err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	return appStructure, rootSrcDir, rootGitpkgDir, nil
}

// getLocalBuildContext retrieves mandatory info to start the build locally:
//   - The target architecture
//   - The target distribution
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param app: Pointer to the full structure of the app to build
// :type app: *rpapiv1.RPApplicationFull
// :returns: string, string, error -- architecture, distribution and an error if a problem occurred
func getLocalBuildContext(cmd *cobra.Command, app *rpapiv1.RPApplicationFull) (string, string, error) {
	archResult, distroResult := "", ""

	// Firstly, get the values given by the user
	// Architecture
	userArch, err := cmd.Flags().GetString(localV2BuildArchFlag)
	if err != nil {
		logFlagError(localV2BuildArchFlag)
		return archResult, distroResult, err
	}
	// Distribution
	userDistribution, err := cmd.Flags().GetString(localV2BuildDistroFlag)
	if err != nil {
		logFlagError(localV2BuildArchFlag)
		return archResult, distroResult, err
	}

	// Check that the distribution given is available for this application
	availableDistros := []string{}
	for _, distro := range app.Distributions {
		if distro.Set && distro.Slug != "" {
			availableDistros = append(availableDistros, distro.Slug)
		}
	}

	// 1st case: no distro specified by the user and more than one distro available
	if (userDistribution == "") && len(availableDistros) > 1 {
		errMsg := fmt.Sprintf(needToSpecifyDistroErr, app.Parent.ProjectSlug)
		errMsg += " (supported distributions: " + strings.Join(availableDistros, ", ") + ")"
		return archResult, distroResult, errors.New(errMsg)
	}

	// 2nd case: distro specified by the user, but not available
	if (userDistribution != "") && !utils.ContainsString(availableDistros, userDistribution) {
		errMsg := fmt.Sprintf(needToChangeDistroErr, userDistribution, app.Parent.ProjectSlug)
		errMsg += " (supported distributions: " + strings.Join(availableDistros, ", ") + ")"
		return archResult, distroResult, errors.New(errMsg)
	}

	// 3rd case: no distribution specified, but only 1 distro available
	if (userDistribution == "") && len(availableDistros) == 1 {
		distroResult = availableDistros[0]
	} else {
		// 4th case: distribution specified, and in available distros
		distroResult = userDistribution
	}

	// Check that the architecture given is available for this project
	// Get the full project
	project := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(app.Parent.ProjectID.String(), &project); err != nil {
		return archResult, distroResult, err
	}
	// Create a slice of the available architecture
	availableArches := []string{}
	for _, arch := range project.Architectures {
		if arch.Set && arch.ID != "" {
			availableArches = append(availableArches, arch.ID)
		}
	}

	// Check if the architecture is available for this project
	if !utils.ContainsString(availableArches, userArch) {
		errMsg := fmt.Sprintf(needToChangeArchErr, userArch, app.Parent.ProjectSlug)
		errMsg += " (supported architectures: " + strings.Join(availableArches, ", ") + ")"
		return archResult, distroResult, errors.New(errMsg)
	} else {
		archResult = userArch
	}

	return archResult, distroResult, nil
}

// fetchGitpkgAndMockFiles can fetch gitpkgs files and the mock configuration files from the redpesk factory,
// and to put those files at the right location
//
// :param localConfig: Configuration to use for the local builder
// :type localConfig: *RPLocalBuildConfigV2
// :param appStructure: Pointer to the application structure
// :type appStructure: *rpapiv1.RPApplicationFull
// :param getGitPkgFiles: Boolean saying that the pkg files need to be retrieved
// :type getGitPkgFiles: bool
// :param getMockFiles: Boolean saying that the mock files need to be retrieved
// :type getMockFiles: bool
// :param force: Ask the function to overwrite the gitpkg directory anyway
// :type force: bool
// :returns: error -- an error if a problem occurred
func fetchGitpkgAndMockFiles(localConfig *RPLocalBuildConfigV2, appStructure *rpapiv1.RPApplicationFull, getGitPkgFiles, getMockFiles, force bool) error {

	// Create local directories for the app if necessary --
	// Display info
	RPPrinter.CustomPrint(localV2FetchCmdAnsLine3)
	// Create the needed repositories if necessary
	err := createLocalBuilderWorkingDirs(localConfig.SourcesDir, localConfig.PackagesDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check gitpkg status for the wanted repo ------------
	fullGitpkgDir := path.Join(localConfig.PackagesDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)
	if !force && getGitPkgFiles {
		// Display info
		RPPrinter.CustomPrint(localV2FetchCmdAnsLine4)
		// If the user does not wan't to force the fetch, check if the md5sum has changed (if some files were already present)
		filesStatus, _, err := checkGitpkgsStatus(fullGitpkgDir, false, true)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
		if !filesStatus {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return errors.New(cannotFetchWithChanges)
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Create the needed variables
	ourCfgPkgDir := path.Join(fullGitpkgDir, localCfgSubDirectory)

	// Fetch the 'gitpkg' files ---------------------------
	if getGitPkgFiles {

		// Try to get the local_builder_configuration.json file if it exists for two reasons:
		//     - Eventually get the branch passed the first time
		//     - Save the configuration if the fetch is done between a "init" and a "build"
		branchConfiguration, localBldConfFile, err := getLocalBuildConfig(ourCfgPkgDir, false)
		if err != nil {
			return err
		}

		// Get the current Active branch
		activeBranch, err := getActiveBranch(appStructure.ID)
		if err != nil {
			return err
		}

		// Now, decide which branch needs to be used, the order of priority is (from more priority to less)
		//    1. User choice through flag
		//    2. Previous user choice through 'local_build_config.json' file
		//    3. Current active branch if both other options are empty
		branchToFetch := ""
		switch {
		case localConfig.Branch != "":
			branchToFetch = localConfig.Branch
		case branchConfiguration.Branch != "":
			branchToFetch = branchConfiguration.Branch
		}
		if branchToFetch != "" {
			// Switch to specified branch if the user gave one
			err = switchToExistingBranch(appStructure.ID, branchToFetch)
			if err != nil {
				return err
			}

			// Defer switching back to current branch
			defer switchToExistingBranch(appStructure.ID, activeBranch)

		} else {
			branchToFetch = activeBranch
		}

		// Save the branch used in the configuration structure
		// It will be later saved in the 'local_build_config.json' file
		branchConfiguration.Branch = branchToFetch

		// Display info
		RPPrinter.CustomPrintF(localV2FetchCmdAnsLine5, branchToFetch)

		// Get the list of all files and directories
		fileDirList, err := getLocalGitpkgsFileOrDirsList(fullGitpkgDir, false)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Delete all the files and then all directories
		for _, tmpFile := range fileDirList["files"] {
			err := os.Remove(tmpFile)
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}
		}
		for _, tmpDir := range fileDirList["dirs"] {
			err := os.Remove(tmpDir)
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}
		}

		// Download the full gitpkg in the gitpkg/ directory
		files := []string{}
		content := rpapiv1.RPFileList{}
		// Ask for the content (the list of files)
		err = bckAPI.ApplicationGetListFiles(appStructure.ID.String(), &content)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
		for _, item := range content.ListFiles {
			files = append(files, item.Name)
		}
		err = downloadGitPkgsAppFiles(appStructure.ID, fullGitpkgDir, files, false)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Create the md5sum for each file
		err = updateOrCreateMD5SumFile(fullGitpkgDir, []string{})
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Write a file in the package directory to save the configuration
		localConfig.Branch = branchConfiguration.Branch
		err = utils.WriteJsonFile(localBldConfFile, *localConfig)
		if err != nil {
			return err
		}

		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Fetch the 'mockconfig' file ------------------------
	if getMockFiles {
		// Retrieve all the build configuration files
		RPPrinter.CustomPrint(localV2FetchCmdAnsLine6)
		err = createRepoInheritanceFile(appStructure.Parent, ourCfgPkgDir)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	return nil
}

// prePackagingRoutineV2 is a common "routine" between `package init` and
// `package build`. It allows to retrieve all the information that are common
// to `run` and `init` that have been set by the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: RPLocalBuildConfigV2, RPApplicationFull, error -- build config, the local app structure
// and an error if something went wrong
func prePackagingRoutineV2(cmd *cobra.Command, args []string) (RPLocalBuildConfigV2, rpapiv1.RPApplicationFull, error) {
	// Initialize the main structures we need to return
	fullConfiguration := RPLocalBuildConfigV2{}

	// Get all the context needed for the packaging
	appStructure, tmpSrcsDir, tmpPkgsDir, err := getLocalV2Context(cmd, args, false)
	if err != nil {
		return fullConfiguration, appStructure, err
	}

	// Save the packages and sources directories
	fullConfiguration.SourcesDir = tmpSrcsDir
	fullConfiguration.PackagesDir = tmpPkgsDir

	// Display the path to both directories here
	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine3, fullConfiguration.SourcesDir)
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine4, fullConfiguration.PackagesDir)
	RPPrinter.CustomPrintLn()

	return fullConfiguration, appStructure, nil
}

// cloneSourcesRepo can be called to clone the sources GIT repository at the
// right place. It also manages the status check to any already existing directory
//
// :param app: Structure containing all needed data for application
// :type app: *rpapiv1.RPApplicationFull
// :param config: Configuration to set for the local builder
// :type config: *RPLocalBuildConfigV2
// :param force: Ask the function to overwrite the gitsource directory anyway
// :type force: bool
// :returns: error -- an error if a problem occurred
func cloneSourcesRepo(app *rpapiv1.RPApplicationFull, localConfig *RPLocalBuildConfigV2, force bool) error {

	// Check gitsources status for the wanted repo --------
	fullGitSourcesDir := path.Join(localConfig.SourcesDir, app.Parent.ProjectSlug, app.PackageName)
	// Check if directory exists and not empty
	isDir, _ := utils.IsDirectory(fullGitSourcesDir)
	isDirEmpty, _ := utils.IsDirectoryEmpty(fullGitSourcesDir)
	if !force && isDir && !isDirEmpty {
		// Display info
		RPPrinter.CustomPrint(localV2PkgInitAnsLine0)
		// If the user does not wan't to force the clone, check if some modifications have been done in the repo
		gitStatusOutput, err := utils.RunCmdInDirectory(fullGitSourcesDir, "git", "status")
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			Log.Warnf(noGitSourceFoundWarn, gitStatusOutput)
		} else if !strings.Contains(gitStatusOutput, "nothing to commit, working tree clean") {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return errors.New(cannotCloneWithChanges)
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Remove the previous source directory if it exists --
	if isDir && !isDirEmpty {
		// Display info
		RPPrinter.CustomPrint(localV2PkgInitAnsLine1)
		// Remove the directory
		err := os.RemoveAll(fullGitSourcesDir)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Finally, clone the source repository ---------------
	// Use `iotpkg init` command
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine7)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	initRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: path.Join(localBuilderGitPkgsDir, app.Parent.ProjectSlug, app.PackageName),
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: []string{"--sourcedir", path.Join(localBuilderGitSourcesDir, app.Parent.ProjectSlug, app.PackageName)},
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgInitPost(&initRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)
	// ----------------------------------------------------

	return err
}

// handleIotpkgSubCmdHelp is used to handle the help requests from the user on
// the following commands
//
// - rp-cli local-v2 package service
// - rp-cli local-v2 package metadata
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param originalUsrArgs: Original arguments given by the user before being filtered
// :type originalUsrArgs: []string
// :param subCmdName: Name of the sub-command to handle
// :type subCmdName: string
// :returns: bool, error -- true if the user asked for help and an error if something went wrong
func handleIotpkgSubCmdHelp(cmd *cobra.Command, originalUsrArgs []string, subCmdName string) (bool, error) {
	// Firstly, check if something has been detected earlier
	if len(originalUsrArgs) == 0 {
		// No help needed
		return false, nil
	}

	ourArgs := originalUsrArgs[0:]

	// Since iotpkg does not like "help", convert every "help" to "--help"
	utils.ReplaceInStringSlice(&ourArgs, "help", "--help")

	// Get the elements that are interesting us
	ourIndex := utils.IndexOf(ourArgs, subCmdName)

	// Get rid of the first two args, to keep only the ones compatible with iotpkg
	iotpkgArgs := ourArgs[ourIndex:]
	subCommandArgs := iotpkgArgs[1:]
	// If the user gave an app name, get rid of it
	if len(iotpkgArgs[1:]) > 2 {
		subCommandArgs = iotpkgArgs[2:]
	}

	// Get the usage from rp-cli
	RPPrinter.CustomPrint(cmd.UsageString())

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()
	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return true, err
	}

	helpRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: "/tmp/",
		},
		SubcmdFlags: subCommandArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	if iotpkgArgs[0] == "service" {
		err = localBAPI.IotpkgServicePost(&helpRequest, &procTicket)
	} else if iotpkgArgs[0] == "metadata" {
		err = localBAPI.IotpkgMetadataPost(&helpRequest, &procTicket)
	} else if iotpkgArgs[0] == "build" {
		err = localBAPI.IotpkgRPBuildPost(&helpRequest, &procTicket)
	}
	if err != nil {
		return true, err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)

	return true, err
}

// getMockNameFromCfgFile is used to extract the true mock name from the configuration
// file used to construct it
//
// :param pathToCfgFile: Path to the iotpkg file to extract the mock name from
// :type pathToCfgFile: string
// :returns: string, error -- name of the corresponding mock and an error if something went wrong
func getMockNameFromCfgFile(pathToCfgFile string) (string, error) {
	// Read the file
	fileContent, err := os.ReadFile(pathToCfgFile)
	if err != nil {
		return "", err
	}

	// Extract the line that is interesting us
	mockNameLine := utils.ExtractLineWithPrefix(string(fileContent), "config_opts['root']")
	if mockNameLine == "" {
		return "", nil
	}

	// Now extract the mock name
	splitLine := strings.Split(mockNameLine, "=")
	if len(splitLine) < 2 {
		return "", nil
	}
	mockName := strings.ReplaceAll(strings.TrimSpace(splitLine[1]), "'", "")

	return mockName, nil
}

// handleIotpkgCleanMock is used to request iotpkg to clean a given mock
//
// :param mockName: Name of the mock to clean
// :type mockName: string
// :returns: bool, error -- true if the user asked for help and an error if something went wrong
func handleIotpkgCleanMock(mockName string) error {
	// Local-builder tag
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	cleanRequest := ladapi.RPLADIotpkgCmdStructV2{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: "/tmp",
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: []string{"-n", mockName},
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgCleanMockPost(&cleanRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localV2LocalBldLogsTag2)

	return err
}

// getIotpkgArgs is a function allowing to filter out the flags and values
// used by rp-cli. In order to work, the following parameter must have been set this way:
//
//	FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: []string, error -- the arguments that need to be passed to iotpkg and an error if a problem appeared
func getIotpkgArgs(cmd *cobra.Command) ([]string, error) {
	returnSlice, itemToDel := []string{}, []string{}

	// Firstly get the args given by the user (get rid of 'rp-cli' name)
	origArgs := os.Args[1:]

	// Retrieve the command path that will obviously appears in the full command-line args
	cmdPath := strings.Split(cmd.CommandPath(), " ")

	// Add the cmd path to the items to delete (but without the rp-cli name though, which is in first position)
	itemToDel = append(itemToDel, cmdPath[1:]...)

	// Now visit all flags
	filterOutRPCliFlags := func(flag *pflag.Flag) {
		// Add the name in our slice if the flag has been passed by the user
		if flag.Changed {
			Log.Debugf("[LOCAL-V2] Filtering out %s flag\n", flag.Name)
			fullFlag := fmt.Sprintf("--%s", flag.Name)
			shortFlag := fmt.Sprintf("-%s", flag.Shorthand)

			// Boolean case - no additional value to delete
			if flag.NoOptDefVal != "" {
				// Delete what exists in the command line
				itemToDel = append(itemToDel, fullFlag, shortFlag)
			} else {
				// Non-boolean case, we need to get rid of the flag and the value!
				// Check if the user use the flag assignment with '='
				fullFlagWithEq := fmt.Sprintf("%s=%s", fullFlag, flag.Value.String())
				shortFlagWithEq := fmt.Sprintf("%s=%s", shortFlag, flag.Value.String())
				if utils.ContainsString(origArgs, fullFlagWithEq) {
					itemToDel = append(itemToDel, fullFlagWithEq)
				} else if utils.ContainsString(origArgs, fullFlagWithEq) {
					itemToDel = append(itemToDel, shortFlagWithEq)
				} else {
					itemToDel = append(itemToDel, fullFlag, shortFlag, flag.Value.String())
				}
			}
		}
	}
	cmd.Flags().VisitAll(filterOutRPCliFlags)

	Log.Debugf("[LOCAL-V2] Items to delete from the original args: %+v\n", itemToDel)

	// Now create arguments for iotpkg
	for _, item := range origArgs {
		if !utils.ContainsString(itemToDel, item) {
			returnSlice = append(returnSlice, item)
		}
	}

	// Get rid of the application slug before returning
	return returnSlice[1:], nil
}

// checkLocalBuilderMountPoints is a function that test if the mount-points from
// the host to the local-builder is correctly done (directories and UID/GID mapping)
//
// :returns: error -- an error if a problem appeared
func checkLocalBuilderMountPoints() error {
	// Retrieve the local-builder directories
	rootSrcDir, rootGitpkgDir, err := getAndChecksBuildDirs("")
	if err != nil {
		return err
	}

	// Create a temporary files in gitsources/ and in gitpkg/ directories
	srcFile, err := os.CreateTemp(rootSrcDir, "")
	if err != nil {
		return err
	}
	defer os.Remove(srcFile.Name())
	srcFileInLocalBuilder := path.Join(localBuilderGitSourcesDir, path.Base(srcFile.Name()))

	pkgFile, err := os.CreateTemp(rootGitpkgDir, "")
	if err != nil {
		return err
	}
	defer os.Remove(pkgFile.Name())
	pkgFileInLocalBuilder := path.Join(localBuilderGitPkgsDir, path.Base(pkgFile.Name()))

	// Check that the files are correctly seen inside the local-builder
	cmdOutput, err := execInLocalBuilderStatic("ls -al "+srcFileInLocalBuilder, "/tmp")
	if err != nil {
		return err
	}
	if !strings.Contains(cmdOutput, "devel mock") {
		return errors.New(pbOnMountPoints)
	}

	cmdOutput, err = execInLocalBuilderStatic("ls -al "+pkgFileInLocalBuilder, "/tmp")
	if err != nil {
		return err
	}
	if !strings.Contains(cmdOutput, "devel mock") {
		return errors.New(pbOnMountPoints)
	}

	return nil
}

// execInLocalBuilderStatic executes a command inside the local-builder and returns
// the output of this command. Nothing is interactive here, the user does not see anything
//
// :param cmdLine: Command line to run in the local-builder
// :type cmdLine: string
// :param workingDir: Working directory where the command needs to be run
// :type workingDir: string
// :returns: error -- an error if a problem appeared
func execInLocalBuilderStatic(cmdLine, workingDir string) (string, error) {

	// Request body
	execRequest := ladapi.RPLADExecCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: 50,
				WinSizeCol:  50,
			},
			WorkingDir: workingDir,
		},
		CmdLine: cmdLine,
	}

	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err := localBAPI.ExecPost(&execRequest, &procTicket)
	if err != nil {
		return "", err
	}

	// Get the output from the redpesk-lad websocket
	// Block until we receive all the websocket events (until localProcEndEvt)
	procRunning := true
	cmdOutput := ""
	var errToReturn error
	for procRunning {
		select {
		// We received an event from the remote server
		case localEvent := <-cmdEventLocalBChannel:
			// Check that the received event is for our process
			if localEvent.ID == procTicket.ID {
				switch localEvent.Type {
				case localStdOutEvt:
					cmdOutput += localEvent.Line
				case localStdErrEvt:
					cmdOutput += localEvent.Line
				case localProcEndEvt:
					// This is the end of the terminal
					errorMsg := fmt.Sprintf(endOfProc, localEvent.ID, localEvent.Status)
					Log.Debugln(errorMsg)
					// End this blocking code
					procRunning = false
					// Return an error if the terminal end badly
					if localEvent.Status != 0 {
						errToReturn = errors.New(errorMsg)
					} else {
						errToReturn = nil
					}
				}
			}
		// Disconnection from the websocket, since everything go through the
		// socket, stop the command
		case <-localBConnection.WSocket.DisconnectionChan:
			// End this blocking code
			procRunning = false
			Log.Error(webSocketDisconnectionErr)
			errToReturn = errors.New(webSocketDisconnectionErr)
		}
	}

	return cmdOutput, errToReturn
}

// checkLocalBuilderMountPoints is a function that test if the qemu
// is correctly set-up on the host machine
//
// :returns: error -- an error if a problem appeared
func checkQemuSetUp() error {
	// Detect the linux distribution
	linuxDistro := utils.GetLinuxDistro()
	if linuxDistro == "ubuntu" {
		// Check partially the content of /var/lib/binfmts/qemu-aarch64
		binFmtAarch64Content, err := os.ReadFile(getQemuAarch64FilePath("/var/lib/binfmts/qemu-aarch64"))
		if err != nil {
			return err
		}
		if !strings.Contains(string(binFmtAarch64Content), varLibBinfmtsQemuAarch64) {
			Log.Error(varLibBinfmtsFileNotGood)
			Log.Errorf("Try to fix this by running the following command:\n%s\n", fixVarLibBinfmtsQemuAarch64file)
			return errors.New(varLibBinfmtsFileNotGood)
		}
	}

	// Check the content of /proc/sys/fs/binfmt_misc/qemu-aarch64
	binFmtMiscAarch64Content, err := os.ReadFile(getQemuAarch64FilePath("/proc/sys/fs/binfmt_misc/qemu-aarch64"))
	if err != nil {
		return err
	}
	if !strings.Contains(string(binFmtMiscAarch64Content), procSysFsBinfmtMiscQemuAarch64V0) &&
		!strings.Contains(string(binFmtMiscAarch64Content), procSysFsBinfmtMiscQemuAarch64V1) {
		return errors.New(procSysFsBinfmtsFileNotGood)
	}
	return nil
}

// getQemuAarch64FilePath is here uniquely for compatibility with
// CI test to redirect to the mock files
//
// :returns: error -- an error if a problem appeared
func getQemuAarch64FilePath(initialPath string) string {
	if runMode == "testing" && strings.Contains(initialPath, "/proc/sys/fs") {
		return utils.GetPkgRootDir() + "rpladmock/binfmt_qemu/proc_sys_fs_binfmt_misc_qemu-aarch64"
	} else if runMode == "testing" && strings.Contains(initialPath, "/var/lib/binfmts") {
		return utils.GetPkgRootDir() + "rpladmock/binfmt_qemu/var_lib_binfmts_qemu-aarch64"
	} else {
		return initialPath
	}
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------
