/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"fmt"
	"os"
	"path"
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"
)

// Constants

// TestLocalV2PackageFetch is the function allowing to test if
// the command "rp-cli local-v2 package fetch" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageFetch(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-fetch-project", "redpesk-lts-arz", "chickenpox", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the gitpkg files of the newly created app
	cmdOutput := checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "chickenpox-0")

	// Check various things on the output string
	assert.Contains(test, cmdOutput, "Retrieve application details")
	assert.Contains(test, cmdOutput, "Check local-builder configuration")
	assert.Contains(test, cmdOutput, "Create directories if needed")
	assert.Contains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository from branch 'main'")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")

	// Check that the right files exist now
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-project/chickenpox-0/chickenpox-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-project/chickenpox-0/chickenpox-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-project/chickenpox-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-project/chickenpox-0/.rp-cli/local-v2-fetch-project.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-project/chickenpox-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-project/chickenpox-0/.rp-cli/local_build_config.json")
}

// TestLocalV2PackageFetchTestFlags is the function allowing to test if
// the command "rp-cli local-v2 package fetch" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageFetchTestFlags(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and applications for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-fetch-flags-project", "redpesk-lts-arz", "chickenpox-flags", 2, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Gitpkg flag --------------------
	cmdOutput := checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "chickenpox-flags-0", "-g")
	// Check various things on the output string
	assert.Contains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository")
	assert.NotContains(test, cmdOutput, "Fetch local-builder configuration")
	// Check that the right files exist now
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-0/chickenpox-flags-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-0/chickenpox-flags-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-0/.rp-cli/local_build_config.json")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-0/.rp-cli/local-v2-fetch-flags-project.json")
	// --------------------------------

	// Localbuilder flag --------------
	cmdOutput = checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "chickenpox-flags-1", "-c")
	// Check various things on the output string
	assert.NotContains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.NotContains(test, cmdOutput, "Fetch the 'gitpkg' repository")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")
	// Check that the right files exist now
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-1/chickenpox-flags-1.spec")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-1/chickenpox-flags-1-1.0.0.tar.gz")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-1/.redpesk/autogen::service.json")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-1/.rp-cli/md5sum.json")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-1/.rp-cli/local_build_config.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-flags-project/chickenpox-flags-1/.rp-cli/local-v2-fetch-flags-project.json")
	// --------------------------------

	// Force flag ---------------------
	cmdOutput = checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "chickenpox-flags-0", "-f")
	// Check various things on the output string
	assert.NotContains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")
	// --------------------------------
}

// TestLocalV2PackageFetchBranchFlag is the function allowing to test if
// the command "rp-cli local-v2 package fetch" is working correctly when
// the "--branch" flag is used
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageFetchBranchFlag(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-fetch-branch-project", "redpesk-lts-arz", "wednesday", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, "local-v2-fetch-branch-project", "wednesday-0", "main", "my-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the gitpkg files of the newly created app, on a given
	cmdOutput := checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "wednesday-0", "--project", "local-v2-fetch-branch-project", "-b", "my-branch")

	// Check various things on the output string
	assert.Contains(test, cmdOutput, "Retrieve application details")
	assert.Contains(test, cmdOutput, "Check local-builder configuration")
	assert.Contains(test, cmdOutput, "Create directories if needed")
	assert.Contains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository from branch 'my-branch'")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")

	// Check that the right files exist now
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-branch-project/wednesday-0/wednesday-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-branch-project/wednesday-0/wednesday-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-branch-project/wednesday-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-branch-project/wednesday-0/.rp-cli/local-v2-fetch-branch-project.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-branch-project/wednesday-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-fetch-branch-project/wednesday-0/.rp-cli/local_build_config.json")

	// Check that the 'local_build_config.json' contains the right branch
	pathToPkgDir := path.Join("/tmp/redpesk-cli-test/gitpkgs", "local-v2-fetch-branch-project", "wednesday-0")
	localBldCfg, _, err := getLocalBuildConfig(pathToPkgDir, true)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "my-branch", localBldCfg.Branch)
}

// TestLocalV2PackageStatus is the function allowing to test if
// the command "rp-cli local-v2 package status" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageStatus(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-status-project", "redpesk-lts-arz", "chickenpox-status", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the files of the newly created app
	checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "chickenpox-status-0")

	// Try a first time to check the status of the "chickenpox-status-0" app
	outputCmd, err := executeCommand(localV2Command, localPkgCmdName, "status", "chickenpox-status-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the status is good
	assert.Contains(test, outputCmd, localV2StatusNoChanges)

	// Now, modify a file, delete another and add another one
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-v2-status-project/chickenpox-status-0/chickenpox-status-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}
	err = os.Remove("/tmp/redpesk-cli-test/gitpkgs/local-v2-status-project/chickenpox-status-0/chickenpox-status-0-1.0.0.tar.gz")
	if err != nil {
		test.Fatal(err)
	}
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-v2-status-project/chickenpox-status-0/new-file.txt", []byte("NEW FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Try another time to check the status of the "chickenpox-status-0" app
	outputCmd, err = executeCommand(localV2Command, localPkgCmdName, "status", "chickenpox-status-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the status is good
	assert.Contains(test, outputCmd, "added:    new-file.txt")
	assert.Contains(test, outputCmd, "deleted:  chickenpox-status-0-1.0.0.tar.gz")
	assert.Contains(test, outputCmd, "modified: chickenpox-status-0.spec")
}

// TestLocalV2PackageCommit is the function allowing to test if
// the command "rp-cli local-v2 package commit" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageCommit(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and applications for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-commit-project", "redpesk-lts-arz", "chickenpox-commit", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the files of the newly created app
	checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "chickenpox-commit-0")

	// Try to commit immediately
	outputCmd, err := executeCommand(localV2Command, localPkgCmdName, "commit", "chickenpox-commit-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that we have nothing to commit
	assert.Contains(test, outputCmd, localV2CommitNothingToCommit)

	// Now, modify a file, delete another and add another one
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-v2-commit-project/chickenpox-commit-0/chickenpox-commit-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}
	err = os.Remove("/tmp/redpesk-cli-test/gitpkgs/local-v2-commit-project/chickenpox-commit-0/chickenpox-commit-0-1.0.0.tar.gz")
	if err != nil {
		test.Fatal(err)
	}
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-v2-commit-project/chickenpox-commit-0/new-file.txt", []byte("NEW FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Now try to commit just one file
	outputCmd, err = executeCommand(localV2Command, localPkgCmdName, "commit", "chickenpox-commit-0", "new-file.txt")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the file has been committed
	assert.Contains(test, outputCmd, "Upload of 'new-file.txt'")
	assert.NotContains(test, outputCmd, "Upload of 'chickenpox-commit-0.spec'")
	assert.NotContains(test, outputCmd, "Removal of 'chickenpox-commit-0-1.0.0.tar.gz'")

	// Then, commit the other files
	outputCmd, err = executeCommand(localV2Command, localPkgCmdName, "commit", "chickenpox-commit-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the files has been committed
	assert.NotContains(test, outputCmd, "Upload of 'new-file.txt'")
	assert.Contains(test, outputCmd, "Upload of 'chickenpox-commit-0.spec'")
	assert.Contains(test, outputCmd, "Removal of 'chickenpox-commit-0-1.0.0.tar.gz'")
}

// TestLocalV2PackageCommitOnBranch is the function allowing to test if
// the command "rp-cli local-v2 package commit" is working correctly when a
// branch has been specified by the user at fetch time
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageCommitOnBranch(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-commit-branch-project", "redpesk-lts-arz", "friday", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, "local-v2-commit-branch-project", "friday-0", "main", "my-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the gitpkg files of the newly created app, on a given
	checkActionCmdAns(test, localV2Command, localPkgCmdName, "fetch", "friday-0", "--project", "local-v2-commit-branch-project", "-b", "my-branch")

	// Now, modify a file
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-v2-commit-branch-project/friday-0/friday-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Then, commit the other files
	outputCmd, err := executeCommand(localV2Command, localPkgCmdName, "commit", "friday-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the file has been committed
	assert.Contains(test, outputCmd, "Upload of 'friday-0.spec'")
}

// TestLocalV2PackageInit is the function allowing to test if
// the command "rp-cli local-v2 package init" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageInit(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-init-project", "redpesk-lts-arz", "lily-pad", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "init", "lily-pad-0", "--project", "local-v2-init-project", "-b", "main")
	if err != nil {
		test.Fatal(err)
	}
	cmdOutput = strings.TrimSpace(cmdOutput)

	// Check the output of the command
	checkPkgV2InitCmd(test, cmdOutput, "lily-pad-0", "main")

	// Check that the right files have been set
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-init-project/lily-pad-0/lily-pad-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-init-project/lily-pad-0/lily-pad-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-init-project/lily-pad-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-init-project/lily-pad-0/.rp-cli/local-v2-init-project.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-init-project/lily-pad-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-v2-init-project/lily-pad-0/.rp-cli/local_build_config.json")

	// Check that the 'local_build_config.json' contains the right values
	pathToPkgDir := path.Join("/tmp/redpesk-cli-test/gitpkgs", "local-v2-init-project", "lily-pad-0")
	localBldCfg, _, err := getLocalBuildConfig(pathToPkgDir, true)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "/tmp/redpesk-cli-test/gitsources", localBldCfg.SourcesDir)
	assert.Equal(test, "/tmp/redpesk-cli-test/gitpkgs", localBldCfg.PackagesDir)
	assert.Equal(test, "main", localBldCfg.Branch)
}

// TestLocalV2PackageService is the function allowing to test if
// the command "rp-cli local-v2 package service" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageService(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-service-project", "redpesk-lts-arz", "service-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm(localV2Command, localPkgCmdName, "init", "service-app-0", "--project", "local-v2-service-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to run a specific service
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "service", "service-app-0", "--project", "local-v2-service-project", "excludearch", "--arches", "[x86_64]")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check the local-builder logs tags
	assert.Contains(test, cmdOutput, localV2LocalBldLogsTag1)
	assert.Contains(test, cmdOutput, localV2LocalBldLogsTag2)
}

// TestLocalV2PackageServiceHelpFlag is the function allowing to test if
// the command "rp-cli local-v2 package service --help" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageServiceHelpFlag(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Ask to run a specific service
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "service", "--help")
	if err != nil {
		test.Fatal(err)
	}

	// Check the local-builder logs tags
	assert.Contains(test, cmdOutput, localV2LocalBldLogsTag1)
	assert.Contains(test, cmdOutput, localV2LocalBldLogsTag2)

	// Extract the logs from mocked repdesk-lad
	localBuilderLogs := extractSubLogsWithFlag(cmdOutput, "LOCAL-BUILDER LOGS")

	// Check that the first line and the last line of the help command is here
	assert.Contains(test, localBuilderLogs, "usage: iotpkg service [-h] [--no-service]")
	assert.Contains(test, localBuilderLogs, "go_dependency       Fetch go dependencies")
}

// TestLocalV2PackageMetadata is the function allowing to test if
// the command "rp-cli local-v2 package metadata" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageMetadata(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-metadata-project", "redpesk-lts-arz", "metadata-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm(localV2Command, localPkgCmdName, "init", "metadata-app-0", "--project", "local-v2-metadata-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to run a specific service
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "metadata", "metadata-app-0", "--project", "local-v2-metadata-project", "--revision", "1.0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check the local-builder logs tags
	assert.Contains(test, cmdOutput, localV2LocalBldLogsTag1)
	assert.Contains(test, cmdOutput, localV2LocalBldLogsTag2)
}

// TestLocalV2PackageBuild is the function allowing to test if
// the command "rp-cli local-v2 package build" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageBuild(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-build-project", "redpesk-lts-arz", "build-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm(localV2Command, localPkgCmdName, "init", "build-app-0", "--project", "local-v2-build-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "build", "build-app-0", "--project", "local-v2-build-project", "--arch", "x86_64", "--distro", "redpesk-lts-arz")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that everything went well
	assert.Contains(test, cmdOutput, localV2BuildRunCmdAnsResultGood)
	assert.Contains(test, cmdOutput, localPkgRunCmdAnsLine4)
}

// TestLocalV2PackageShell is the function allowing to test if
// the command "rp-cli local-v2 package shell" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageShell(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-shell-project", "redpesk-lts-arz", "shell-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm(localV2Command, localPkgCmdName, "init", "shell-app-0", "--project", "local-v2-shell-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to get a shell
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "shell", "shell-app-0", "--project", "local-v2-shell-project", "--arch", "aarch64", "--distro", "redpesk-lts-arz")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that we got a pseudo prompt
	assert.Contains(test, cmdOutput, "devel@rp-localbuilder")
}

// TestLocalV2PackageClean is the function allowing to test if
// the command "rp-cli local-v2 package clean" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2PackageClean(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-v2-clean-project", "redpesk-lts-arz", "clean-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm(localV2Command, localPkgCmdName, "init", "clean-app-0", "--project", "local-v2-clean-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	_, err = execCmdInSimTerm(localV2Command, localPkgCmdName, "build", "clean-app-0", "--project", "local-v2-clean-project")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the right directories exist
	assert.DirExists(test, path.Join(gitPackagesTestDir, "local-v2-clean-project", "clean-app-0", ".iotpkg"))
	assert.FileExists(test, path.Join(gitPackagesTestDir, "local-v2-clean-project", "clean-app-0", ".iotpkg", "clean-app-0"+"-redpesk-distro-aarch64-cross.cfg"))
	assert.FileExists(test, path.Join(gitPackagesTestDir, "clean-app-0"+"-redpesk-distro-aarch64-cross.cfg"))

	// Try to clean
	cmdOutput, err := execCmdInSimTerm(localV2Command, localPkgCmdName, "clean", "clean-app-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check the logs
	assert.Contains(test, cmdOutput, localPkgCleanCmdAnsLine2)
	assert.Contains(test, cmdOutput, "Start: clean chroot")
	assert.Contains(test, cmdOutput, "Finish: clean chroot")

	// Check that the directories have been deleted
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-v2-clean-project"))
	assert.NoFileExists(test, path.Join(gitPackagesTestDir, "clean-app-0"+"-redpesk-distro-aarch64-cross.cfg"))
}

// TestLocalV2CheckSetup is the function allowing to test if
// the command "rp-cli local-v2 check-setup" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalV2CheckSetup(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, localV2Command, localV2CheckSetupCmd)
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// checkPkgV2InitCmd is the function allowing to check the output of
// the command "rp-cli local-V2 package init ..."
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param cmdOutput : Output from the rp-cli local package init
// :type cmdOutput: string
func checkPkgV2InitCmd(test *testing.T, cmdOutput, appName, branchName string) {

	// Check that the directories used are the one we used for tests
	checkLineEndWith(test, cmdOutput, "\"gitsources/\" directory used:", gitSourcesTestDir)
	checkLineEndWith(test, cmdOutput, "\"gitpkgs/\" directory used:", gitPackagesTestDir)

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	checkLineEndWith(test, cmdOutput, fmt.Sprintf(localPkgInitCmdAnsLine5, appName), "[OK]")
	checkLineEndWith(test, cmdOutput, "Create directories if needed", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check 'gitpkg' repository status", "[OK]")
	checkLineEndWith(test, cmdOutput, fmt.Sprintf("Fetch the 'gitpkg' repository from branch '%s'", branchName), "[OK]")
	checkLineEndWith(test, cmdOutput, "Fetch local-builder configuration", "[OK]")

	// Extract the clone logs
	cloneLogs := extractSubLogsWithFlag(cmdOutput, "LOCAL-BUILDER LOGS")

	// Check that there are some clone logs and that they look coherent
	assert.NotEqual(test, "", cloneLogs)
	assert.Contains(test, cloneLogs, "Successfull shallow cloned")
}
