/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"encoding/json"
	"fmt"
	"os"
	"path"
	"strconv"
	"strings"
	"testing"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpconnect"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rputils"
	homedir "github.com/mitchellh/go-homedir"
	"github.com/stretchr/testify/assert"
)

// Constants
const (
	// Directories used for tests
	gitSourcesTestDir  = "/tmp/redpesk-cli-test/gitsources"
	gitPackagesTestDir = "/tmp/redpesk-cli-test/gitpkgs"
	cfgDirPathTestDir  = "/tmp/redpesk-cli-test/.redpesk/rp-cli/"

	// Package parameters used for testing
	package4Tests              = "redtest-helloworld-api"
	package4TestsURL           = "http://git.ovh.iot/redpesk/redpesk-samples/redtest-helloworld-api.git"
	package4TestsSpecPathInSrc = "./conf.d/packaging/rpm/redtest-helloworld-api.spec"
	package4TestsSpecPathURL   = "http://git.ovh.iot/redpesk/redpesk-samples/redtest-helloworld-api/-/raw/master/conf.d/packaging/rpm/redtest-helloworld-api.spec"
	package4TestsSpecPathLocal = "/tmp/redpesk-cli-test/redtest-helloworld-api.spec"

	package4TestsHelloAlias     = "my-helloworld-binding"
	package4TestsHelloName      = "My Helloworld binding"
	package4TestsHelloURL       = "http://git.ovh.iot/redpesk/redpesk-samples/helloworld-binding.git"
	package4TestsHelloPathInSrc = "conf.d/packaging/helloworld-binding.spec"

	gitpkgPackage4TestsURL  = "http://git.ovh.iot/incubator/gitpkgd-helloworld-gitpkg.git"
	gitpkgPackage4TestsName = "gitpkgd-helloworld-gitpkg"

	albatrosPoem = `L’albatros

Souvent, pour s’amuser, les hommes d’équipage
Prennent des albatros, vastes oiseaux des mers,
Qui suivent, indolents compagnons de voyage,
Le navire glissant sur les gouffres amers.

A peine les ont-ils déposés sur les planches,
Que ces rois de l’azur, maladroits et honteux,
Laissent piteusement leurs grandes ailes blanches
Comme des avirons traîner à côté d’eux.

Ce voyageur ailé, comme il est gauche et veule !
Lui, naguère si beau, qu’il est comique et laid !
L’un agace son bec avec un brûle-gueule,
L’autre mime, en boitant, l’infirme qui volait !

Le Poète est semblable au prince des nuées
Qui hante la tempête et se rit de l’archer ;
Exilé sur le sol au milieu des huées,
Ses ailes de géant l’empêchent de marcher.

Charles Baudelaire`

	// Connections
	localConnectDisplayNoVerboseColNb = 2
	localConnectDisplayVerboseLineNb  = 2
)

// TestOnBoardLocalBuilder is the function allowing to test if
// the command "rp-cli onboard" is working correctly for the
// local-builder connection part
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestOnBoardLocalBuilder(test *testing.T) {

	// We will have to fake the stdin in this situation (because this command is interactive)
	// See https://eli.thegreenplace.net/2020/faking-stdin-and-stdout-in-go/ for more info

	// Create a false local_builder configuration file
	cfgFilePath, err := createFalseLocalConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Get the home directory
	home, err := homedir.Dir()
	if err != nil {
		test.Fatal(err)
	}
	gitsourcesPath := path.Join(home, "my_rp_builder_dir", "gitsources")

	// Check if the directory exists
	defBuildDirExists := rputils.Exists(gitsourcesPath)

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(cfgFilePath)

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will fill what the command 'rp-cli onboard' wants
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we do not want to configure the backend connection
		writePipe.Write([]byte("n\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we want to configure the local-builder connection
		writePipe.Write([]byte("Y\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe the localbuilder IP
		writePipe.Write([]byte("http://0.0.0.0:8112\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe that we don't want to modify the IP
		writePipe.Write([]byte("n\n"))

		if defBuildDirExists {
			time.Sleep(100 * time.Millisecond)
			// Write in the pipe that we don't want to modify the IP
			writePipe.Write([]byte("n\n"))
		}

		// Write in the pipe
		time.Sleep(100 * time.Millisecond)
		writePipe.Write([]byte("/tmp/my_rp_builder_dir/gitsources\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we want to create the directory
		writePipe.Write([]byte("Y\n"))

		if defBuildDirExists {
			time.Sleep(100 * time.Millisecond)
			// Write in the pipe that we don't want to modify the IP
			writePipe.Write([]byte("n\n"))
		}

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe
		writePipe.Write([]byte("/tmp/my_rp_builder_dir/gitpkgs\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we want to create the directory
		writePipe.Write([]byte("Y\n"))
	}()

	// Execute the onboard command
	cmdOutput, err := executeCommand(rootOnBoardCmd)

	// Restore Stdin
	os.Stdin = origStdin

	if err != nil {
		test.Fatal(err)
	}

	// Remove the directories that have been created
	defer os.RemoveAll("/tmp/my_rp_builder_dir")

	// Check that the new value in the local builder configuration file is good
	writenConnection, err := getConnectionParams(localBuilderAlias, localBuilderFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "http://0.0.0.0:8112", writenConnection.ServerURL)

	// Check that in cmdOutput, an [OK] appears
	checkLineEndWith(test, cmdOutput, "> Modification of the local builder file...", "[OK]")

	// Check that the directories have been created
	assert.DirExists(test, "/tmp/my_rp_builder_dir/gitsources")
	assert.DirExists(test, "/tmp/my_rp_builder_dir/gitpkgs")
}

// TestLocalPkgInitSpecInSrc is the function allowing to test if
// the command "rp-cli local package init" is working correctly
// when the spec file is among the sources
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgInitSpecInSrc(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Yuksek project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "yuksek-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathInSrc, "yuksek-project", false, false, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"yuksek-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgInitSpecInSrcOffline is the function allowing to test if
// the command "rp-cli local package init" is working correctly
// when the spec file is among the sources and when the user request an "offline" packaging
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgInitSpecInSrcOffline(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Yuksek project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "yuksek-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathInSrc, "yuksek-project", false, false, true)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"yuksek-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Check that all "build configuration files" have been downloaded
	_, err = checkAllBuildConfigFiles(test, path.Join(gitPackagesTestDir, "yuksek-project", package4Tests), []string{"redpesk-devel-arz"}, []string{"aarch64", "x86_64"})
	if err != nil {
		test.Fatal(err)
	}

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgInitSpecOnWeb is the function allowing to test if
// the command "rp-cli local package init" is working correctly
// when the spec file needs to be downloaded from the web.
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgInitSpecOnWeb(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Gesleir project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "gesleir-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathURL, "gesleir-project", false, false, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"gesleir-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgInitSpecLocalOffline is the function allowing to test if
// the command "rp-cli local package init" is working correctly
// when the spec file is local and when the user request an "offline" packaging
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgInitSpecLocalOffline(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Download the spec file in order to have it locally
	err = utils.HTTPDwnldFileInDir("/tmp/redpesk-cli-test", package4TestsSpecPathURL)
	if err != nil {
		test.Fatal(err)
	}

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("JukeBoyz project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "jukeboyz-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathLocal, "jukeboyz-project", true, false, true)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"jukeboyz-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Check that the "build configuration files" have been all downloaded
	_, err = checkAllBuildConfigFiles(test, path.Join(gitPackagesTestDir, "jukeboyz-project", package4Tests), []string{"redpesk-devel-arz"}, []string{"aarch64", "x86_64"})
	if err != nil {
		test.Fatal(err)
	}

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgInitOnlyPkg is the function allowing to test if
// the command "rp-cli local package init" is working correctly
// when the user has access only to the "git package"
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgInitOnlyPkg(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Akhenaton project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "akhenaton-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(gitpkgPackage4TestsName, gitpkgPackage4TestsURL, "", "akhenaton-project", false, true, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, false, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"akhenaton-project"+"/"+gitpkgPackage4TestsName+"/SOURCES/agl-service-helloworld-1.0.0.tar.gz")
	assert.FileExists(test, gitPackagesTestDir+"/"+"akhenaton-project"+"/"+gitpkgPackage4TestsName+"/SPECS/agl-service-helloworld.spec")

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgInitAppExistInBck is the function allowing to test if
// the command "rp-cli local package init" is working correctly
// when the application is already existing in the backend
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgInitAppExistInBck(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Dopamine project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "dopamine-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an "helloworld-binding" application
	cmdToRun4App := getMainAddAppCmdWithSpec(package4TestsHelloName, package4TestsHelloAlias, "dopamine-project",
		package4TestsHelloURL,
		package4TestsHelloPathInSrc)

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, package4TestsHelloAlias, cmdToRun4App...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageAppExistInitCmd(package4TestsHelloAlias, "dopamine-project")

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/dopamine-project/my-helloworld-binding/my-helloworld-binding-1.0.1.tar.gz")

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgRunSpecInSrcX86_64 is the function allowing to test if
// the command "rp-cli local package build" is working correctly
// when the spec file is among the sources and on a x86_64
// architecture
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgRunSpecInSrcX86_64(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Yuksek project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "yuksek-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathInSrc, "yuksek-project", false, false, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"yuksek-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Create the command to run the `rp-cli local package build` command
	cmdToRun4PkgRun := getPackageRunCmd(package4Tests, "yuksek-project", "x86_64", false)

	// Run the command to run the packaging and check the output of it
	runAndCheckPkgRunCmd(test, "yuksek-project", cmdToRun4PkgRun...)

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgRunSpecInSrcX86_64Offline is the function allowing to test if
// the command "rp-cli local package build" is working correctly
// when the spec file is among the sources and on a x86_64
// architecture and the user requests an offline packaging
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgRunSpecInSrcX86_64Offline(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Yuksek project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "yuksek-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathInSrc, "yuksek-project", false, false, true)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"yuksek-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Create the command to run the `rp-cli local package build` command
	cmdToRun4PkgRun := getPackageRunCmd(package4Tests, "yuksek-project", "x86_64", false)

	// Run the command to run the packaging and check the output of it
	runAndCheckPkgRunCmd(test, "yuksek-project", cmdToRun4PkgRun...)

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgRunSpecInSrcX86_64RegenTarBall is the function allowing to test if
// the command "rp-cli local package build" is working correctly
// when the spec file is among the sources and on a x86_64
// architecture and that the the tarball needs to be regenerated
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgRunSpecInSrcX86_64RegenTarBall(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Yuksek project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "yuksek-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathInSrc, "yuksek-project", false, false, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	fullTarBallPath := gitPackagesTestDir + "/" + "yuksek-project" + "/" + package4Tests + "/redtest-helloworld-api-1.0.tar.gz"
	tarballExists := assert.FileExists(test, fullTarBallPath)
	if tarballExists {
		// Delete the tarball in order to request for the regeneration afterward
		err = os.Remove(fullTarBallPath)
		if err != nil {
			test.Fatal(err)
		}
	}

	// Create the command to run the `rp-cli local package build` command
	cmdToRun4PkgRun := getPackageRunCmd(package4Tests, "yuksek-project", "x86_64", true)

	// Run the command to run the packaging and check the output of it
	runAndCheckPkgRunCmd(test, "yuksek-project", cmdToRun4PkgRun...)

	// Check that the tarball has been regenerated
	assert.FileExists(test, fullTarBallPath)

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgRunSpecOnWebAarch64 is the function allowing to test if
// the command "rp-cli local package build" is working correctly
// when the spec file needs to be downloaded from the web and on a aarch64
// architecture
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgRunSpecOnWebAarch64(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Gesleir project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "gesleir-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathURL, "gesleir-project", false, false, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"gesleir-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Create the command to run the `rp-cli local package build` command
	cmdToRun4PkgRun := getPackageRunCmd(package4Tests, "gesleir-project", "aarch64", false)

	// Run the command to run the packaging and check the output of it
	runAndCheckPkgRunCmd(test, "gesleir-project", cmdToRun4PkgRun...)

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgRunSpecOnWebAarch64OfflineSeveralDistros is the function allowing to test if
// the command "rp-cli local package build" is working correctly
// when the spec file needs to be downloaded from the web and on a aarch64
// architecture. The packaging needs to be offline and the project has several enabled distros
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgRunSpecOnWebAarch64OfflineSeveralDistros(test *testing.T) {

	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Worakls project", "redpesk-devel-arz")
	// Add another distribution
	cmdToRun4Prj = append(cmdToRun4Prj, "--optional-distro", "redpesk-devel-33")
	err = addObjectIfNotExists(test, prjsCommand, "worakls-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathURL, "worakls-project", false, false, true)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"worakls-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Create the command to run the `rp-cli local package build` command
	cmdToRun4PkgRun := getPackageRunCmd(package4Tests, "worakls-project", "aarch64", false)

	// Specify the redpesk-devel-arz distro
	cmdToRun4PkgRun = append(cmdToRun4PkgRun, "-d", "redpesk-devel-arz")

	// Run the command to run the packaging and check the output of it
	runAndCheckPkgRunCmd(test, "worakls-project", cmdToRun4PkgRun...)

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgClean is the function allowing to test if
// the command "rp-cli local package clean" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgClean(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-pkg-clean-prj-0", "redpesk-lts-arz", "local-pkg-clean-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create several directories and sub-directories to simulate app and projects
	err = createFakeBldDirs(1, 1, "local-pkg-clean", false)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the directory we want to delete exists
	assert.DirExists(test, path.Join(gitPackagesTestDir, "local-pkg-clean-prj-0", "local-pkg-clean-app-0"))

	// Run the command
	output, err := executeCommand(localCommand, localPkgCmdName, "clean", "local-pkg-clean-app-0", "--project", "local-pkg-clean-prj-0")
	if err != nil {
		test.Fatal(err)
	}
	assert.Contains(test, output, localPkgCleanCmdAnsLine2)

	// Check that the directory does not exist anymore
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-pkg-clean-prj-0", "local-pkg-clean-app-0"))

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgCleanSrcModified is the function allowing to test if
// the command "rp-cli local package clean" is working correctly
// even when the sources have been modified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgCleanSrcModified(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-pkg-clean-src-prj-0", "redpesk-lts-arz", "local-pkg-clean-src-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create several directories and sub-directories to simulate app and projects
	err = createFakeBldDirs(1, 1, "local-pkg-clean-src", true)
	if err != nil {
		test.Fatal(err)
	}

	// Now clone a simple project in one of the created source repository
	prjSrcDirectory := path.Join(gitSourcesTestDir, "local-pkg-clean-src-prj-0")
	_, err = utils.RunCmdInDirectory(prjSrcDirectory, "git", "clone", package4TestsHelloURL, "local-pkg-clean-src-app-0")
	if err != nil {
		test.Fatal(err)
	}

	// Modify something in the just cloned repo
	appSrcDirectory := path.Join(prjSrcDirectory, "local-pkg-clean-src-app-0")
	_, err = utils.RunCmdInDirectory(appSrcDirectory, "touch", "modification.txt")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to delete the directory

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will answer the request of the command rp-cli local package clean
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we still want to delete the directory
		writePipe.Write([]byte("y\n"))
	}()

	// Run the command
	output, err := executeCommand(localCommand, localPkgCmdName, "clean", "local-pkg-clean-src-app-0")

	// Restore Stdin
	os.Stdin = origStdin

	if err != nil {
		test.Fatal(err)
	}

	// Check that the output contains a request from rp-cli
	assert.Contains(test, output, localPkgCleanCmdAnsLine1)
	assert.Contains(test, output, localPkgCleanCmdAnsLine2)

	// Check that the directory does not exist anymore
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-pkg-clean-src-prj-0", "local-pkg-clean-src-app-0"))

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgCleanSrcModified is the function allowing to test if
// the command "rp-cli local package clean -f" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgCleanForce(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-pkg-clean-force-prj-0", "redpesk-lts-arz", "local-pkg-clean-force-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create several directories and sub-directories to simulate app and projects
	err = createFakeBldDirs(1, 1, "local-pkg-clean-force", true)
	if err != nil {
		test.Fatal(err)
	}

	// Now clone a simple project in one of the created source repository
	prjSrcDirectory := path.Join(gitSourcesTestDir, "local-pkg-clean-force-prj-0")
	_, err = utils.RunCmdInDirectory(prjSrcDirectory, "git", "clone", package4TestsHelloURL, "local-pkg-clean-force-app-0")
	if err != nil {
		test.Fatal(err)
	}

	// Modify something in the just cloned repo
	appSrcDirectory := path.Join(prjSrcDirectory, "local-pkg-clean-force-app-0")
	_, err = utils.RunCmdInDirectory(appSrcDirectory, "touch", "modification.txt")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to delete the directory

	// Run the command
	output, err := executeCommand(localCommand, localPkgCmdName, "clean", "local-pkg-clean-force-app-0", "--force")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the output contains a request from rp-cli
	assert.Contains(test, output, localPkgCleanCmdAnsLine2)

	// Check that the directory does not exist anymore
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-pkg-clean-force-prj-0", "local-pkg-clean-force-app-0"))

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgCleanInit is the function allowing to test if
// the command "rp-cli local package clean --init" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgCleanInit(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project in the backend (necessary to initialize a package)
	cmdToRun4Prj := getMainAddPrjCmd("Yuksek project", "redpesk-devel-arz")
	err = addObjectIfNotExists(test, prjsCommand, "yuksek-project", cmdToRun4Prj...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an app in the backend
	cmdToRun := getMainAddAppCmd(package4Tests, package4Tests, "yuksek-project")
	err = addObjectIfNotExists(test, appsCommand, package4Tests, cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to run
	cmdToRun4PkgInit := getPackageInitCmd(package4Tests, package4TestsURL, package4TestsSpecPathInSrc, "yuksek-project", false, false, false)

	// Run the command to initialize the packaging and check the output of it
	runAndCheckPkgInitCmd(test, true, cmdToRun4PkgInit...)

	// Check that the tar.gz has been created
	assert.FileExists(test, gitPackagesTestDir+"/"+"yuksek-project"+"/"+package4Tests+"/redtest-helloworld-api-1.0.tar.gz")

	// Now try to clean the project with the init flag
	output, err := execCmdInSimTerm(localCommand, localPkgCmdName, "clean", "redtest-helloworld-api", "-i")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the output contains a request from rp-cli
	assert.Contains(test, output, localPkgCleanCmdAnsLine2)

	// Extract the "local package init" part
	outputTab := strings.Split(output, localPkgInitCmdAnsLine0)
	if assert.Equal(test, 2, len(outputTab)) {
		checkPkgInitCmd(test, true, outputTab[1])
	}

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalExecCmdLs is the function allowing to test if
// the command "rp-cli local exec" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalExecCmdLs(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a directory in our gitsources/ used for test
	testDir := gitSourcesTestDir + "/test-local-exec-dir"
	err := utils.CreateDirIfNotExists(testDir, true)
	if err != nil {
		test.Fatal(err)
	}

	// Create few tests files
	utils.WriteToFile(testDir+"/test_01.log", "THIS IS A FIRST FILE!!!")
	utils.WriteToFile(testDir+"/test_iot_bzh.log", "THIS IS SPARTA!!!")
	utils.WriteToFile(testDir+"/redpesk_test_lorient.log", "TOUS A LA TAVERNE!!!\n LOVE")
	utils.WriteToFile(testDir+"/linux_i_love_you.log", "THIS HURTS HERE, RIGHT IN MY MEOW MEOW")

	// Run the "ls -al" command
	cmdOutput, err := execCmdInSimTerm(localCommand, "exec", "-w", testDir, "--", "ls", "-ahl")
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	// Check that the returned output contains all the file created
	assert.Contains(test, cmdOutput, "test_01.log")
	assert.Contains(test, cmdOutput, "test_iot_bzh.log")
	assert.Contains(test, cmdOutput, "redpesk_test_lorient.log")
	assert.Contains(test, cmdOutput, "linux_i_love_you.log")
}

// TestLocalExecCmdSmartWDLink is the function allowing to test if
// the command "rp-cli local exec" is working correctly when launched
// from a gitpkgs/ or gitsources/ directory
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalExecCmdSmartWDLink(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a directory in our gitsources/ used for test
	testDir := gitSourcesTestDir + "/test-local-exec-smart-link"
	err = utils.CreateDirIfNotExists(testDir, true)
	if err != nil {
		test.Fatal(err)
	}

	// Create few tests files
	utils.WriteToFile(testDir+"/redpesk_test_lorient.log", "TOUS A LA TAVERNE!!!\n LOVE")
	utils.WriteToFile(testDir+"/linux_i_love_you.log", "THIS HURTS HERE, RIGHT IN MY MEOW MEOW")

	// Change the current directory from Golang point of view
	err = os.Chdir(testDir)
	if err != nil {
		test.Fatal(err)
	}

	// Run the "ls -al" command
	cmdOutput, err := execCmdInSimTerm(localCommand, "exec", "--", "ls", "-ahl")
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	// Check that the returned output contains all the file created
	assert.Contains(test, cmdOutput, "redpesk_test_lorient.log")
	assert.Contains(test, cmdOutput, "linux_i_love_you.log")
}

// TestLocalExecCmdEchoStdin is the function allowing to test if
// the command "rp-cli local exec" is working correctly with a bit
// of interactivity using the stdin event
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalExecCmdEchoStdin(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a directory in our gitsources/ used for test
	testDir := gitSourcesTestDir + "/test-local-exec-dir"
	err := utils.CreateDirIfNotExists(testDir, true)
	if err != nil {
		test.Fatal(err)
	}

	// Define the stdin to pass to the command
	stdinString := albatrosPoem

	// Calculate the md5sum on the stdin test
	stdinMd5Sum := utils.CalculateMd5SumOnString(stdinString)

	// Define the variables that we will use for test retries...
	testCounter := 0
	testOK := false
	md5SumCreatedFile := ""
	for !testOK && testCounter < 3 {
		// Write some logs
		test.Logf("[INFO] Attempt n°%d to validate stdin test", testCounter)

		// Run the "cat" command in an interactive way
		_, err = execCmdInSimTermWithStdin(stdinString, localCommand, "exec", "-w", testDir, "--", "sleep", "5")
		if err != nil {
			// The command returns an error
			test.Fatal(err)
		}

		// Check that the file has been created and contains what we write in it
		filesTab, err := utils.FindFilesWithExtension(testDir, ".log")
		assert.Nil(test, err)
		assert.Len(test, filesTab, 1)
		assert.Regexp(test, "test_file_.*.log$", filesTab[0])

		// Calculate the md5sum on the created file
		md5SumCreatedFile, err = utils.CalculateMd5SumOnFile(filesTab[0], true)
		if err != nil {
			test.Fatal(err)
		}

		if md5SumCreatedFile == stdinMd5Sum {
			testOK = true
		}

		// Increase the counter
		testCounter += 1

		// Remove the file
		err = os.Remove(filesTab[0])
		if err != nil {
			test.Fatal(err)
		}
	}

	// Test that the md5sums are the same
	assert.Equal(test, stdinMd5Sum, md5SumCreatedFile)
}

// TestLocalConnectionsAdd is the function allowing to test if
// the command "rp-cli local connections add" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsAdd(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-testadd",
		ServerURL: "10.231.232.5",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the connection has been correctly added
	regexpMap := map[string]string{
		"Alias": `^localbuilder-testadd$`,
		"IP":    `^10\.231\.232\.5$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "get", "localbuilder-testadd")
}

// TestLocalConnectionsGet is the function allowing to test if
// the command "rp-cli local connections get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsGet(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-testget",
		ServerURL: "10.231.232.6",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the connection has been correctly added
	regexpMap := map[string]string{
		"Alias": `^localbuilder-testget$`,
		"IP":    `^10\.231\.232\.6$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "get", "localbuilder-testget")

	// Modify the map in order to test the verbose mode
	regexpMap["Local-builder IP"] = `^10\.231\.232\.6$`

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, localConnectDisplayVerboseLineNb, "", regexpMap, localCommand, connectionsCommand, "get", "localbuilder-testget", "-v")
}

// TestLocalConnectionsUseAndCurrent is the function allowing to test if
// the commands "rp-cli local connections use" and
// "rp-cli local connections current" are working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsUseAndCurrent(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-use-current-0",
		ServerURL: "10.231.232.7",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create another connection
	connectToCreate = rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-use-current-1",
		ServerURL: "10.231.232.8",
	}
	_, err = addLocalConnectionForTest(test, &connectToCreate)
	if err != nil {
		test.Fatal(err)
	}

	// Request rp-cli to use a connection that we just add
	_, err = executeCommand(localCommand, connectionsCommand, "use", "localbuilder-test-use-current-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the current connection returned is "localbuilder-test-use-current-0"
	regexpMap := map[string]string{
		"Alias": `^localbuilder-test-use-current-0$`,
		"IP":    `^10\.231\.232\.7$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "current")

	// Verbose mode
	regexpMap["Local-builder IP"] = `^10\.231\.232\.7$`
	checkCmdAnswerListKeysRegexp(test, localConnectDisplayVerboseLineNb, "", regexpMap, localCommand, connectionsCommand, "current", "-v")
}

// TestLocalConnectionsList is the function allowing to test if
// the command "rp-cli local connections list" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsList(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-list-0",
		ServerURL: "10.231.233.1",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create other connections
	for i := 1; i <= 3; i++ {
		connectToCreate.Alias = fmt.Sprintf("localbuilder-test-list-%d", i)
		_, err = addLocalConnectionForTest(test, &connectToCreate)
		if err != nil {
			test.Fatal(err)
		}
	}

	// Check now that the connections created are correctly listed

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Alias": `^localbuilder-test-list-[0-9]{1}$`,
		"IP":    `^10\.231\.233\.1$`,
	}
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "list")

	// Verbose mode
	regexpMap["Local-builder IP"] = `^10\.231\.233\.1$`
	checkCmdAnswerListKeysRegexp(test, localConnectDisplayVerboseLineNb, "", regexpMap, localCommand, connectionsCommand, "list", "-v")
}

// TestLocalConnectionsUpdate is the function allowing to test if
// the command "rp-cli local connections update" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsUpdate(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-update",
		ServerURL: "10.231.234.1",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, localCommand, connectionsCommand, "update", "localbuilder-test-update", "-i", "10.231.234.2")

	// Check that the update command correctly worked
	regexpMap := map[string]string{
		"Alias": `^localbuilder-test-update$`,
		"IP":    `^10\.231\.234\.2$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "get", "localbuilder-test-update")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// emptyLocalTestDirs allows to empty the directories necessary
// to test the "local" subcommand (gitpkgs and gitsources)
//
// :returns: error -- an error if a problem has occurred
func emptyLocalTestDirs() error {
	// Empty the gitsources directory
	err := utils.RemoveDirContents(gitSourcesTestDir)
	if err != nil {
		return err
	}

	// Empty the gitpkgs directory
	err = utils.RemoveDirContents(gitPackagesTestDir)
	if err != nil {
		return err
	}

	return nil
}

// getPackageInitCmd is an initializer function to get the mainly
// used command to initializing the packaging
//
// :param name: Name of the package to init
// :type name: string
// :param gitURL: URL to the GIT repository
// :type gitURL: string
// :param specPath: Path to the spec file: locally, in an URL or relative in GIT
// :type specPath: string
// :param prjAlias: Alias of the parent project for the app
// :type prjAlias: string
// :param isSpecLocal: To set to true if the spec file is local
// :type isSpecLocal: bool
// :param isGitPkg: To set to true if the git repos is a gitpkg
// :type isGitPkg: bool
// :param offline: Specify if the packaging then needs to be offline
// :type offline: bool
// :returns: []string -- the value of the wanted constant
func getPackageInitCmd(name, gitURL, specPath, prjAlias string, isSpecLocal, isGitPkg, offline bool) []string {
	// Define the spec flag to use
	specFlagToUse := "--spec-path"
	if isSpecLocal {
		specFlagToUse = "--spec-local"
	}

	// Define the URL flag to use
	urlFlagToUse := "--source-url"
	if isGitPkg {
		urlFlagToUse = "--package-url"
	}

	result := []string{
		localCommand,
		localPkgCmdName,
		"init",
		name,
		"--project",
		prjAlias,
		urlFlagToUse,
		gitURL,
		"-P",
		gitPackagesTestDir,
		"-S",
		gitSourcesTestDir,
	}

	if !isGitPkg {
		result = append(result, specFlagToUse, specPath)
	}

	if offline {
		result = append(result, "--offline")
	}

	return result
}

// getPackageAppExistInitCmd is an initializer function to get the mainly
// used command to initializing the packaging
//
// :param name: Name of the package to init
// :type name: string
// :param prjAlias: Alias of the parent project for the app
// :type prjAlias: string
// :returns: []string -- the value of the wanted constant
func getPackageAppExistInitCmd(name, prjAlias string) []string {

	result := []string{
		localCommand,
		localPkgCmdName,
		"init",
		name,
		"--project",
		prjAlias,
		"-P",
		gitPackagesTestDir,
		"-S",
		gitSourcesTestDir,
	}

	return result
}

// getPackageRunCmd is an initializer function to get the mainly
// used command to run the packaging
//
// :param name: Name of the package to init
// :type name: string
// :param prjAlias: Alias of the parent project for the app
// :type prjAlias: string
// :param arch: Architecture to package for
// :type arch: string
// :param regenTarball: Request to regenerate the tarball
// :type regenTarball: bool
// :returns: []string -- the value of the wanted constant
func getPackageRunCmd(name, prjAlias, arch string, regenTarball bool) []string {

	result := []string{
		localCommand,
		localPkgCmdName,
		"build",
		name,
		"--project",
		prjAlias,
		"--target",
		arch,
		"-P",
		gitPackagesTestDir,
		"-S",
		gitSourcesTestDir,
	}

	if regenTarball {
		result = append(result, "--regen-archive")
	}

	return result
}

// checkLineEndWith allows to check that, in a given text, there is a line
// that starts with the given prefix and that also finish with the given
// suffix
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param fullText: Full text where the line is hidden
// :type fullText: string
// :param prefix: Prefix of the line to look for
// :type prefix: string
// :param suffix: Suffix that the line needs to contain
// :type suffix: string
func checkLineEndWith(test *testing.T, fullText, prefix, suffix string) {
	// Get the line we want
	lineWanted := utils.ExtractLineWithPrefix(fullText, prefix)

	// Check that the line has been found
	assert.NotEqual(test, "", lineWanted)

	// Check that this line ends with what the caller wants
	assert.True(test, strings.HasSuffix(lineWanted, suffix))
}

// runAndCheckPkgInitCmd is the function allowing to run and check the output of
// the command "rp-cli local package init ..."
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param checkArchiveLogs: To set to true if the archive logs needs to be checked
// :type checkArchiveLogs: bool
// :param args: Args to pass as command to create the project
// :type args: ...string
func runAndCheckPkgInitCmd(test *testing.T, checkArchiveLogs bool, args ...string) {
	// Run the commands in a simulated terminal
	cmdOutput, err := execCmdInSimTerm(args...)
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	checkPkgInitCmd(test, checkArchiveLogs, cmdOutput)
}

// checkPkgInitCmd is the function allowing to check the output of
// the command "rp-cli local package init ..."
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param checkArchiveLogs: To set to true if the archive logs needs to be checked
// :type checkArchiveLogs: bool
// :param cmdOutput : Output from the rp-cli local package init
// :type cmdOutput: string
func checkPkgInitCmd(test *testing.T, checkArchiveLogs bool, cmdOutput string) {

	// Check that the directories used are the one we used for tests
	checkLineEndWith(test, cmdOutput, "\"gitsources/\" directory used:", gitSourcesTestDir)
	checkLineEndWith(test, cmdOutput, "\"gitpkgs/\" directory used:", gitPackagesTestDir)

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Checking application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Checking packaging directories...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Retrieving package configuration for", "[OK]")

	// Extract the clone logs
	cloneLogs := extractSubLogsWithFlag(cmdOutput, "CLONE LOGS")

	// Check that there are some clone logs and that they look coherent
	assert.NotEqual(test, "", cloneLogs)
	assert.True(test, strings.HasPrefix(cloneLogs, "Cloning into"))

	if checkArchiveLogs {
		// Extract the archive logs
		archiveLogs := extractSubLogsWithFlag(cmdOutput, "ARCHIVE LOGS")

		// Check that there are some archive logs and that they are coherent
		assert.NotEqual(test, "", archiveLogs)
		archiveLogsFile := utils.GetPkgRootDir() + "rpladmock/logs/iotpkg-archive.log"
		md5SumArchLogsFile, err := utils.CalculateMd5SumOnFile(archiveLogsFile, true)
		if err != nil {
			test.Fatal(err)
		}
		md5SumArchLogsStr := utils.CalculateMd5SumOnString(archiveLogs)

		archLogsEqual := assert.Equal(test, md5SumArchLogsFile, md5SumArchLogsStr)
		if !archLogsEqual {
			fmt.Printf("Archive logs from string: \n%s\n", archiveLogs)
		}
	}
}

// runAndCheckPkgRunCmd is the function allowing to run and check the output of
// the command "rp-cli local package build ..."
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjSlug: Slug of the project containing the app to build
// :type prjSlug: string
// :param args: Args to pass as command to create the project
// :type args: ...string
func runAndCheckPkgRunCmd(test *testing.T, prjSlug string, args ...string) {
	// Run the commands in a simulated terminal
	cmdOutput, err := execCmdInSimTerm(args...)
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	// Default log file is for cross-build
	logFile := "rpladmock/logs/iotpkg-localcrossbuild-helloworld.log"
	if utils.ContainsString(args, "x86_64") {
		logFile = "rpladmock/logs/iotpkg-localbuild-helloworld.log"
	}

	// Get package name
	pkgName := args[3]

	// Check that the directories used are the one we used for tests
	checkLineEndWith(test, cmdOutput, "\"gitsources/\" directory used:", gitSourcesTestDir+"/"+prjSlug+"/"+pkgName)
	checkLineEndWith(test, cmdOutput, "\"gitpkgs/\" directory used:", gitPackagesTestDir+"/"+prjSlug+"/"+pkgName)

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Checking application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Checking packaging directories...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Retrieving packaging parameters for", "[OK]")

	// Check that the logs contains the fact that the packaging went well
	assert.Contains(test, cmdOutput, "Results and/or logs in:")
	assert.Contains(test, cmdOutput, "Congratulations! Your packaging seems to be successful!")

	// Extract the archive logs
	packagingLogs := extractSubLogsWithFlag(cmdOutput, "PACKAGING LOGS")

	// Check that there are some archive logs and that they are coherent
	assert.NotEqual(test, "", packagingLogs)
	packagingLogsFile := utils.GetPkgRootDir() + logFile
	md5SumArchLogsFile, err := utils.CalculateMd5SumOnFile(packagingLogsFile, true)
	if err != nil {
		test.Fatal(err)
	}
	md5SumArchLogsStr := utils.CalculateMd5SumOnString(packagingLogs)

	archLogsEqual := assert.Equal(test, md5SumArchLogsFile, md5SumArchLogsStr)
	if !archLogsEqual {
		fmt.Printf("Archive logs from string: \n%s\n", packagingLogs)
	}
}

// extractSubLogsWithFlag is the function allowing to extract some sublogs
// separated from the main logs with flags looking like:
//
// - Start flag: "CLONE LOGS >>>"
// - End flag: "<<< CLONE LOGS"
//
// :param fullLogs: Full logs where there are the sublogs to extract
// :type fullLogs: string
// :param flag: Flag to look for in order to extract the sublogs
// :type flag: string
// :returns: string -- the sublogs wanted
func extractSubLogsWithFlag(fullLogs, flag string) string {
	// Construct the flags (begin and end) from the flag given by the caller
	startFlag := flag + " >>>"
	endFlag := "<<< " + flag

	// First, check that both flags are inside
	if !strings.Contains(fullLogs, startFlag) || !strings.Contains(fullLogs, endFlag) {
		return ""
	}

	// Split the full logs
	tmpSlice := strings.Split(fullLogs, startFlag)
	if len(tmpSlice) != 2 {
		return ""
	}

	// Split the first substring
	tmpSlice2 := strings.Split(tmpSlice[1], endFlag)
	if len(tmpSlice2) != 2 {
		return ""
	}

	finalSubLogs := strings.TrimSpace(tmpSlice2[0])

	return finalSubLogs
}

// checkAllBuildConfigFiles is the function allowing to check that all the configuration files
// necessary for the local build are correctly downloaded in the right directory
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param directory: Directory where the files need to be
// :type directory: string
// :param allDistros: Slice of all distributions available
// :type allDistros: []string
// :param allArches: Slice of all architectures available
// :type allArches: []string
func checkAllBuildConfigFiles(test *testing.T, directory string, allDistros, allArches []string) (bool, error) {
	// Initialize result
	result := false

	// Get the list of all the files that are interesting us in the directory
	buildConfigFileList, err := utils.FindFilesWithExtension(directory, ".cfg")
	if err != nil {
		return result, err
	}

	// Check that, at least we have the right number of files
	result = assert.Equal(test, len(buildConfigFileList), (len(allDistros) * len(allArches)))
	if !result {
		return result, nil
	}

	// Now, for each combination, check that the configuration file has been downloaded
	for _, distro := range allDistros {
		for _, arch := range allArches {
			regularExpressionWanted := fmt.Sprintf(`_%s_%s.cfg$`, distro, arch)
			tmpResult, _, err := utils.RegexMatchNumberInSlice(regularExpressionWanted, buildConfigFileList)
			if err != nil {
				return result, err
			}
			if !assert.GreaterOrEqual(test, tmpResult, 1) {
				return result, nil
			}
		}
	}
	return true, nil
}

// createFalseLocalBuilderFile allows to create a false local-builder file while
// saving the current one thanks to a backup
//
// :returns: string, error -- The path to the file and an error if something went wrong
func createFalseLocalBuilderFile() (string, error) {
	// Get path to the configuration directory
	configDirPath, err := getConfigDir()
	if err != nil {
		return "", err
	}
	localBuilderFilePath := path.Join(configDirPath, localBuilderFile)

	// Convert the connections slice to json string
	connectsJsonBytes, err := json.MarshalIndent([]rpconnect.RPCLIConnection{}, "", "  ")
	if err != nil {
		return "", err
	}

	err = createFalseFile(localBuilderFilePath, string(connectsJsonBytes))
	if err != nil {
		return "", err
	}

	return localBuilderFilePath, nil
}

// addLocalConnectionForTest is a utility function allowing to create a
// new connection and to create a fake local-builder file if it does not
// exist already.
// It returns a path to the fake file created. The calling
// function needs to call 'defer deleteFalseFile(fakeFilePath)'
// in order to delete it at the end of tests
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param connection: Connection to add to rp-cli
// :type connection: *rpconnect.RPCLIConnection
// :returns: string, error -- the path to the fake token file and an error if something bad happened
func addLocalConnectionForTest(test *testing.T, connection *rpconnect.RPCLIConnection) (string, error) {
	// Check if the local-builder file is already existing, and create it if it is not the case
	lbFileExists, lbFilePath, err := fakeFileExists(localBuilderFile)
	if err != nil {
		return lbFilePath, err
	}
	if !lbFileExists {
		lbFilePath, err = createFalseLocalBuilderFile()
		if err != nil {
			return lbFilePath, err
		}
	}

	// Now add the connection object if it does not already exists
	addCmd := []string{localCommand, connectionsCommand, "add", connection.Alias, "-i", connection.ServerURL}

	err = addObjectIfNotExists(test, localCommand+" "+connectionsCommand, connection.Alias, addCmd...)
	if err != nil {
		return lbFilePath, err
	}

	return lbFilePath, nil
}

// createFakeBldDirs is a utility function that creates a given number
// of "project" and "app" directories as they would be created if
// they were created using the "local package init" command
//
// :param prjsNb: Number of project directories to create
// :type prjsNb: int
// :param appsNb: Number of app directories to create inisde each project directory
// :type appsNb: int
// :param namePrefix: Prefix name for each directory
// :type namePrefix: string
// :param createSrcDir: Set to true to also create the source directories
// :type createSrcDir: bool
// :returns: error -- an error if something bad happened
func createFakeBldDirs(prjsNb, appsNb int, namePrefix string, createSrcDir bool) error {
	// Create several directories and sub-directories to simulate app and projects
	for prjIndex := 0; prjIndex < prjsNb; prjIndex++ {
		prjName := namePrefix + "-prj-" + strconv.Itoa(prjIndex)
		prjDirToCreatePkg := path.Join(gitPackagesTestDir, prjName)
		prjDirToCreateSrc := path.Join(gitSourcesTestDir, prjName)

		// Now loop on the different applications
		for appIndex := 0; appIndex < appsNb; appIndex++ {
			appName := namePrefix + "-app-" + strconv.Itoa(appIndex)
			appDirToCreatePkg := path.Join(prjDirToCreatePkg, appName)
			appDirToCreateSrc := path.Join(prjDirToCreateSrc, appName)

			// Create the directories
			err := os.MkdirAll(appDirToCreatePkg, 0755)
			if err != nil {
				return err
			}
			if createSrcDir {
				err = os.MkdirAll(appDirToCreateSrc, 0755)
				if err != nil {
					return err
				}
			}
		}
	}
	return nil
}
